# -*- coding: utf-8 -*-
"""
Created on Fri Apr  5 09:44:14 2024

@author: mfratki
"""
import numpy as np
import pandas as pd
try:
    import baseflow as bf
except:
    pass

'''
monthly
  cfs - mean, median, std
  in -  sum
  mg/l - mean, median, std
  lb - sum

annual







'''
# @dataclass
# class Timeseries:
    
    
# @dataclass
# class HydroData:
    
#     sim
#     obs
#     units
    



# class Metrics():
    
#     data
    




def aggregate(df,units):
    assert units in ['lb','mg/l','in','cfs','degC']
    if units in ['mg/l','cfs','degC']:
        agg_func = 'mean'
    else:
        agg_func = 'sum'
        
    df_agg = pd.DataFrame(np.ones((12,3))*np.nan,index = range(1,13),columns = ['simulated','observed','ratio'])
    df_agg.index.name = 'month'
    df = df.groupby(df.index.month).agg(agg_func)[['simulated','observed']]
    df.columns = ['simulated','observed']
    df['ratio'] = df['observed']/df['simulated']
    df_agg.loc[df.index,df.columns] = df.values
    
    df_agg.loc['Mean'] = df_agg.agg('mean')
    df_agg['ratio'] = df_agg['observed']/df_agg['simulated']
    return df_agg

def NSE(df_daily,agg_func = 'mean'):
    #df_monthly = df_daily.resample('MS').sum()
    df_monthly = df_daily.groupby(df_daily.index.month).agg(agg_func)
    
    NSE_daily = round(1 - np.sum((df_daily['observed'] - df_daily['simulated'])**2)/np.sum((df_daily['observed'] - df_daily['observed'].mean())**2),3)
    NSE_garrick = round(1 - np.sum(np.absolute(df_daily['observed'] - df_daily['simulated']))/np.sum(np.absolute(df_daily['observed'] - df_daily['observed'].mean())),3)
    NSE_monthly = round(1 - np.sum((df_monthly['observed'] - df_monthly['simulated'])**2)/np.sum((df_monthly['observed'] - df_monthly['observed'].mean())**2),3)
    metric = pd.DataFrame(data = [NSE_daily, NSE_garrick, NSE_monthly]).transpose()
    metric.columns = ['NSE','Garrick','Monthly']
    return metric


#%% Hydrology
def hydro_stats(df_daily,drg_area): 
    '''
    Parameters
    ----------
    df_daily : TYPE
        Daily flow timeseries in cfs. A dataframe with a datetime index and columns labeled 'observed','simulated'
    drg_area : TYPE
        drainage area in acres.

    Returns
    -------
    df : TYPE
        Dataframe of various metrics used to evaluate model hydrology simulations.
        Outputs are in inches.

    '''
    df_daily = hydro_sep(drg_area*0.0015625,df_daily) # in cfs
    df_daily_inches = df_daily*60*60*24*12/(drg_area*43560) # in inches
    dfs = [total(df_daily_inches,'in'),
           annual(df_daily_inches,'in'),
           monthly(df_daily_inches,'in'),
           season(df_daily_inches,'in'),
           low_50(df_daily_inches),
           high_10(df_daily_inches),
           storm(df_daily_inches),
           storm_summer(df_daily_inches),
           baseflow(df_daily_inches)]
    
    
    dfs[0] ['interval'] = 'Total'
    dfs[1] ['interval'] = 'Annual'
    dfs[2].rename(columns = {'month':'interval'},inplace = True)
    dfs[3].rename(columns = {'season':'interval'},inplace = True)
    dfs[4] ['interval'] = 'Low 50'
    dfs[5] ['interval'] = 'High 10'
    dfs[6] ['interval'] = 'Storm'
    dfs[7] ['interval'] = 'Summer Storm'
    dfs[8] ['interval'] = 'Baseflow'
    
    
    df = pd.concat(dfs)
    df.set_index('interval',inplace=True)
    nse = NSE(df_daily)
    df.loc['Daily NSE',:] = pd.NA
    df.loc['Daily Garrick',:] = pd.NA
    df.loc['Monthly NSE',:] = pd.NA
    df.loc['Daily NSE','observed'] = nse['NSE'].values
    df.loc['Daily Garrick','observed'] = nse['Garrick'].values
    df.loc['Monthly NSE','observed'] = nse['Monthly'].values
    df.loc['Daily NSE','goal'] = .7
    df.loc['Daily Garrick','goal'] = .55
    df.loc['Monthly NSE','goal'] = .8
    return df




def hydro_sep_baseflow(df_daily,drng_area = None,method = 'Boughton'):
    #dfs,df_kge = bf.single(df_daily['observed'],area = drg_area)
    #df = bf.separation(df_daily['observed'],method = 'Boughton')
    #method = df_kge.idxmax()

    
    df_daily['observed_baseflow'] = bf.single(df_daily['observed'],area = drng_area, method = method,return_kge = False)[0][method]
    df_daily['simulated_baseflow'] = bf.single(df_daily['simulated'],area = drng_area, method = method,return_kge = False)[0][method]
    df_daily['observed_runoff'] = df_daily['observed'] - df_daily['observed_baseflow']
    df_daily['simulated_runoff'] = df_daily['simulated'] - df_daily['simulated_baseflow']
    return df_daily

    
    

def hydro_sep(drg_area,df_daily):
    try:
        df_daily = hydro_sep_baseflow(df_daily,drg_area)
    except:
        twonstar = 2*drg_area**0.2
        twonstar = round(twonstar,0)
        if twonstar%2==0:
            twonstar = twonstar - 1
        twonstar = np.median([3,11,twonstar])
        ndays = (twonstar - 1)/2
        ndays = int(ndays)
    
        df_daily['observed_baseflow'] = df_daily['observed'].rolling(2*ndays+1,center=True,min_periods=ndays+1).apply(np.nanmin)
        df_daily['observed_runoff'] = df_daily['observed'] - df_daily['observed_baseflow']
        df_daily['simulated_baseflow'] = df_daily['simulated'].rolling(2*ndays+1,center=True,min_periods=ndays+1).apply(np.nanmin)
        df_daily['simulated_runoff'] = df_daily['simulated'] - df_daily['simulated_baseflow']
    return df_daily

# def aggregate(df,period = None,agg_func = 'mean'):
#     if periond is None:
#         df = df.agg(agg_func)
#     elif period == 'Y':
#         grouper = df.index.year
#     elif period == 'M':
#         grouper = df.index.month
#     elif period == 'D':
#         grouper = df.index.dayofyear
        
    
def annual(df_daily,units):
    assert units in ['cfs','in','lb','mg/l','degC']
    if units in ['in','lb']:    
        # Total flow volume
        n_years = df_daily.groupby(df_daily.index.year).count()/365.25
        df_daily = ((df_daily.groupby(df_daily.index.year).sum())/n_years).mean()# (df_daily.groupby(df_daily.index.year).count()/365.25) .(agg_func)
    else:
        df_daily = (df_daily.groupby(df_daily.index.year).mean()).mean()# (df_daily.groupby(df_daily.index.year).count()/365.25) .(agg_func)

    
    observed = df_daily['observed']
    simulated = df_daily['simulated']
    error = round((simulated - observed)/observed*100,2)
    observed = float(round(observed,2))
    simulated = float(round(simulated,2))
    metric = { 'observed': [observed], 
             'simulated': [simulated], 
             'per_error': [error], 
             'abs_error': [simulated-observed],
             'goal': [10]}
    return pd.DataFrame(metric)

def total(df_daily,units):
    if units in ['in','lb']:
        agg_func = 'sum'
    else:
        agg_func = 'mean'
    # Calculate monthly flow in inches
    #df_monthly = df_daily.resample('MS').sum()
    #df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).mean()
    #df_monthly_grouped = df_daily.groupby(df_daily.index.month).agg(agg_func)

    # Total flow volume
    df_daily = df_daily.agg(agg_func)
    observed = df_daily['observed']
    simulated = df_daily['simulated']
    error = round((simulated - observed)/observed*100,2)
    observed = float(round(observed,2))
    simulated = float(round(simulated,2))
    metric = { 'observed': [observed], 
             'simulated': [simulated], 
             'per_error': [error], 
             'abs_error': [simulated-observed],
             'goal': [10]}
    return pd.DataFrame(metric)

def monthly(df_daily,units):
    assert units in ['cfs','in','lb','mg/l','degC']
    #df_monthly = df_daily.resample('MS').sum()
    #df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).agg(agg_func)
    n_months = df_daily.groupby(df_daily.index.month).count()/30  
    if units in ['in','lb']:
        df_monthly_grouped = df_daily.groupby(df_daily.index.month).sum()/n_months #.agg(agg_func)
    else:
        df_monthly_grouped = df_daily.groupby(df_daily.index.month).mean()#.agg(agg_func)

    # Monthly average flow volumes
    mont_names = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'] 
    metrics = []
    for month in range(1,13):
        if month in df_monthly_grouped.index:
            observed = df_monthly_grouped.loc[month,'observed']
            simulated = df_monthly_grouped.loc[month,'simulated']
            error = round((simulated-observed)/observed*100,2)
            metric = {'month' : mont_names[month-1],
                     'observed': [observed], 
                     'simulated': [simulated], 
                     'per_error': [error], 
                     'abs_error': [simulated-observed],
                     'goal': [30]}
            
        else:
            metric = {'month' : mont_names[month-1],
                     'observed': [pd.NA], 
                     'simulated': [pd.NA], 
                     'per_error': [pd.NA], 
                     'abs_error': [pd.NA],
                     'goal': [pd.NA]}
        
        metrics.append(pd.DataFrame(metric))
    return pd.concat(metrics).reset_index(drop=True)


def exceedence(df):
        
    df['simulated_rank'] = df['simulated'].rank(method = 'average', ascending = False)
    df['simulated_exceed'] = df['simulated_rank'] / (len(df) + 1) * 100
    df['observed_rank'] = df['observed'].rank(method = 'average', ascending = False)
    df['observed_exceed'] = df['observed_rank'] / (len(df) + 1) * 100
    
    if 'simulated_flow' in df.columns:
        df['simulated_flow_rank'] = df['simulated_flow'].rank(method = 'average', ascending = False)
        df['simulated_flow_exceed'] = df['simulated_flow_rank'] / (len(df) + 1) * 100
        df['observed_flow_rank'] = df['observed_flow'].rank(method = 'average', ascending = False)
        df['observed_flow_exceed'] = df['observed_flow_rank'] / (len(df) + 1) * 100
    

def low_50(df_daily):

     num_years = len(df_daily.index)/365.25
     num_bottom50 =  int(len(df_daily.index)/2)
     # 50% lowest flow volume
     observed = df_daily.sort_values('observed',ascending=False).tail(num_bottom50).sum()/num_years
     simulated = df_daily.sort_values('simulated',ascending=False).tail(num_bottom50).sum()/num_years
     error = round((simulated['simulated'] - observed['observed'])/observed['observed']*100,2)
     observed = float(round(observed['observed'],2))
     simulated = float(round(simulated['simulated'],2))
     metric = { 'observed': [observed], 
              'simulated': [simulated], 
              'per_error': [error], 
              'abs_error': [simulated-observed],
              'goal': [10]}
     return pd.DataFrame(metric)

def high_10(df_daily):
     num_years = len(df_daily.index)/365.25
     num_top10 =  int(len(df_daily.index)/10)
     # 10% highest flow volume
     observed = df_daily.sort_values('observed',ascending=False).head(num_top10).sum()/num_years
     simulated = df_daily.sort_values('simulated',ascending=False).head(num_top10).sum()/num_years
     error = round((simulated['simulated'] - observed['observed'])/observed['observed']*100,2)
     observed = float(round(observed['observed'],2))
     simulated = float(round(simulated['simulated'],2))
     metric = { 'observed': [observed], 
              'simulated': [simulated], 
              'per_error': [error], 
              'abs_error': [simulated-observed],
              'goal': [15]}

     return pd.DataFrame(metric)
    
    

def season(df_daily,units):
    assert units in ['cfs','in','lb','mg/l','degC']

    # Calculate monthly flow in inches
    #df_monthly = df_daily.resample('MS').sum()
    #df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).mean()
    #df_monthly_grouped = df_daily.groupby(df_daily.index.month).agg(agg_func)
    n_months = df_daily.groupby(df_daily.index.month).count()/30  
    
    if units in ['in','lb']:
        agg_func = 'sum'
        df_monthly_grouped = df_daily.groupby(df_daily.index.month).sum()/n_months #.agg(agg_func)
    else:
        agg_func = 'mean'
        df_monthly_grouped = df_daily.groupby(df_daily.index.month).mean() #.agg(agg_func)


    metrics = []
    for season_start in [7,10,1,4]:
        #summer,fall,winter,spring
        # define seasons to allow for descriptive output
        if season_start == 7:
            season = 'summer '
        elif season_start == 10:
            season = 'fall'
        elif season_start == 1:
            season = 'winter'
        elif season_start == 4:
            season = 'spring'
        observed = df_monthly_grouped['observed'].loc[season_start:season_start + 2].agg(agg_func)
        simulated = df_monthly_grouped['simulated'].loc[season_start:season_start + 2].agg(agg_func)
        error = round((simulated - observed)/observed*100,2)
        observed = float(round(observed,2))
        simulated = float(round(simulated,2))
        metrics.append({'season': season,
                  'observed': observed, 
                  'simulated': simulated, 
                  'per_error': error, 
                  'abs_error': simulated-observed,
                  'goal': [30]}) 
    return pd.DataFrame(metrics).reset_index(drop=True)


def storm(df_daily,agg_func = 'mean'):
    # # Calculate monthly flow in inches
    # df_monthly = df_daily.resample('MS').sum()
    # df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).mean()
    
    # # Storm flow volume
    # observed = df_monthly_grouped['observed_runoff'].sum()
    # simulated = df_monthly_grouped['simulated_runoff'].sum()
    
    #df_daily = df_daily.resample('Y').sum().mean()
    #df_daily = df_daily.groupby('Y').agg(agg_func)
    num_years = len(df_daily.index)/365.25
    df_daily = df_daily.sum()/num_years
    observed = df_daily['observed_runoff']
    simulated = df_daily['simulated_runoff']
    
    error = round((simulated - observed)/observed*100,2)
    observed = float(round(observed,2))
    simulated = float(round(simulated,2))
    metric = { 'observed': [observed], 
             'simulated': [simulated], 
             'per_error': [error], 
             'abs_error': [simulated-observed],
             'goal': [20]}
    
    return pd.DataFrame(metric)

def storm_summer(df_daily,agg_func = 'mean'):
    # average annual summer volume
    # Calculate monthly flow in inches
    #df_monthly = df_daily.resample('MS').sum()
    #df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).mean()
    n_months = df_daily.groupby(df_daily.index.month).count()/30  
    df_monthly_grouped = df_daily.groupby(df_daily.index.month).sum()/n_months #.agg(agg_func)


    # Summer storm flow volume
    observed = df_monthly_grouped['observed_runoff'].loc[7:9].sum()
    simulated = df_monthly_grouped['simulated_runoff'].loc[7:9].sum()
    error = round((simulated - observed)/observed*100,2)
    observed = float(round(observed,2))
    simulated = float(round(simulated,2))
    metric = { 'observed': [observed], 
             'simulated': [simulated], 
             'per_error': [error], 
             'abs_error': [simulated-observed],
             'goal': [50]}
    
    return pd.DataFrame(metric)    

def baseflow(df_daily,agg_func = 'mean'):
    # average annual baseflow
    num_years = len(df_daily.index)/365.25
    #df_monthly = df_daily.resample('MS').sum()
    #df_monthly_grouped = df_monthly.groupby(df_monthly.index.month).mean()
    #df_monthly_grouped = df_daily.groupby(df_daily.index.month).agg(agg_func)

    observed = df_daily['observed_baseflow'].sum()/num_years
    simulated = df_daily['simulated_baseflow'].sum()/num_years
    error = round((simulated - observed)/observed*100,2)
    observed = float(round(observed,2))
    simulated = float(round(simulated,2))
    metric = { 'observed': [observed], 
             'simulated': [simulated], 
             'per_error': [error], 
             'abs_error': [simulated-observed],
             'goal': [20]}
    
    return pd.DataFrame(metric)    




#%% Sediment

# def sed_stats(df,units):
#     if units == 'mg/l':
#         agg_func = 'mean'
#     else:
#         agg_func = 'sum'
        
#     df_agg = pd.DataFrame(np.ones((12,3))*np.nan,index = range(1,13),columns = ['model','flux','ratio'])
#     df_agg.index.name = 'month'
#     df = df.groupby(df.index.month).agg(agg_func)[['Simflow','Obsflow']]
#     df.columns = ['model','flux']
#     df['ratio'] = df['flux']/df['model']
#     df_agg.loc[df.index,df.columns] = df.values
    
#     df_agg.loc['Mean'] = df_agg.agg('mean')
#     df_agg['ratio'] = df_agg['flux']/df_agg['model']
#     return df_agg


def stats(df_daily,units):
    dfs = [total(df_daily,units),
           annual(df_daily,units),
           monthly(df_daily,units),
           season(df_daily,units)]
        
 
    dfs[0] ['interval'] = 'Total'
    dfs[1] ['interval'] = 'Annual'
    dfs[2].rename(columns = {'month':'interval'},inplace = True)
    dfs[3].rename(columns = {'season':'interval'},inplace = True)
    df = pd.concat(dfs)
    df.set_index('interval',inplace=True)

    return df


#%% Nutrients
def nutrient_stats(df_daily,units):
    dfs = [total(df_daily,units),
           annual(df_daily,units),
           monthly(df_daily,units),
           season(df_daily,units)]
        
 
    dfs[0] ['interval'] = 'Total'
    dfs[1] ['interval'] = 'Annual'
    dfs[2].rename(columns = {'month':'interval'},inplace = True)
    dfs[3].rename(columns = {'season':'interval'},inplace = True)
    df = pd.concat(dfs)
    df.set_index('interval',inplace=True)

    return df



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 16:51:50 2021

@author: mfratki
"""
#standard imports
from copy import deepcopy
import subprocess
#non-standard imports
import pandas as pd
pd.set_option('display.max_columns', None)
pd.set_option('display.max_rows', None)
# to reset this
pd.reset_option('display.max_columns')
import numpy as np
from pathlib import Path

#My packages
from pyhspf.hspfModel import hspfModel
from pyhspf.wdm import wdmInterface
from pyhspf import helpers
from mpcaHydro import data_manager as dm
from pyhcal import metrics
from pyhcal import figures
from pyhcal.setup_utils import Builder
#from hspf_tools.orm.monitoring_db import MonitoringDatabase


class calProject():
    #valid_models = Builder.valid_models()
    def __init__(self,project_location):
        self.project_location = Path(project_location)
        
    
    def new_project(self,model_name):
        return Builder(model_name) #self._builder.new_project(project_location,model_name)
    
    def load_project(self,model_name):
        if model_name in [f.name for f in self.project_location.iterdir() if f.is_dir()]:
            return calibrator(self.project_location.joinpath(model_name))
        else:
            answer = input("No calibration project for that model. Would you like to set on up? (yes or no")
            if answer.lower() in ["y","yes"]:
                self.new_project(model_name)
                return calibrator(self.project_location.joinpath(model_name))
            elif answer.lower() in ["n","no"]:
                return
                # Do other stuff
            else:
                print('please enter yes or no')
            
    
def config_info(project_folder):
    project_path = Path(project_folder)
    info = {'project_path'  : project_path,
             'project_name' : project_path.name,
             'model_path'   : project_path.joinpath('model'),
             'output_path'  : project_path.joinpath('output'),
             'start_date'   : '1996-01-01',
             'end_date'     : '2100-01-01',
            }     
    return info



        
class calibrator:
    def __init__(self,project_folder):
        self.project_path = Path(project_folder)
        self.project_name = self.project_path.name
        self.model_path = self.project_path.joinpath('model')
        self.output_path = self.project_path.joinpath('output')
        self.run = None
        #self.winHSPF = str(Path(__file__).resolve().parent.parent) + '\\bin\\WinHSPFLt\\WinHspfLt.exe'     
        self.start_date = '1996-01-01'
        self.end_date = '2100-01-01'
        
        # Load observational data into memory TODO: Convert to database?
        self.dm = dm.dataManager(self.project_path.joinpath('data'))
        #self.odm = MonitoringDatabase(cal.project_path.joinpath(cal.project_name))
     
        self.targets = None
        if self.project_path.joinpath('targets.csv').exists():
            self.targets = pd.read_csv(self.project_path.joinpath('targets.csv'))  
              
        self.MODL_DB = pd.read_csv(self.project_path.joinpath('_'.join([self.project_name ,'MODL_DB.csv'])))
        
        self.model = None
        self._wdms = None
        self.uci = None
        
    ## Input/Output methods
    def initialize(self,reach_ids,default = 4):

        self.uci.update_table(default,'RCHRES','BINARY-INFO',0,columns = ['HEATPR','HYDRPR','SEDPR','OXRXPR','NUTRPR','PLNKPR'],operator = 'set')
        self.uci.update_table(2,'RCHRES','BINARY-INFO',0,columns = ['HEATPR','HYDRPR','SEDPR','OXRXPR','NUTRPR','PLNKPR'],opnids = reach_ids,operator = 'set')
        
        self.uci.write(self.model.uci_file)
        winHSPF = str(Path(__file__).resolve().parent.parent) + '\\bin\\WinHSPFLt\\WinHspfLt.exe'     
        subprocess.run([winHSPF,self.model.uci_file]) #, stdout=subprocess.PIPE, creationflags=0x08000000)
        
    def set_dates(self, start_date = '1996-01-01',end_date ='2100-01-01'):
        self.start_date = start_date
        self.end_date = end_date

        
    def load_model(self,name):

        if isinstance(name,int): # Default approach
            self.run = name
            name = '_'.join([self.project_name , str(name)])   
        else:
            name = str(name)
            self.run = None
            
        self.model = hspfModel(self.model_path.joinpath(name + '.uci'))  
        
        # WDM data never changes so I am trying to avoid reloading it each time the model is updated.
        # Very brittle solution. TODO: think harder about how to handle this.
        if self._wdms is None:
            try:
                self._wdms = wdmInterface(self.model.wdm_paths)
            except:
                pass
                
        self.model.wdms = self._wdms
        self.model.reports.wdms = self._wdms
        self.uci = deepcopy(self.model.uci) #uci to be manipulated 

        
    # def setup_run(self, reach_ids = None, time_Step = 3,n = 1):
    #     setup_utils.setup(self.uci,self.project_name,run = self.run,reach_ids = reach_ids,n = 1,time_step = 3)
    
    def run_model(self,name = None,overwrite_hbn = False): # NO STATE CHANGE 
        
        if name is None:
            name = '_'.join([self.project_name, str(self.run+1)])   
        elif isinstance(name,int): # Default approach
             name = '_'.join([self.project_name , str(name)])  
        else:
            name = str(name) 
        
        if not overwrite_hbn:
            self.uci.update_bino(name)
        
        uci_file = self.model_path.joinpath(name + '.uci').as_posix()        
        self.uci.write(uci_file)
        winHSPF = str(Path(__file__).resolve().parent.parent) + '\\bin\\WinHSPFLt\\WinHspfLt.exe'     
        subprocess.run([winHSPF,uci_file]) #, stdout=subprocess.PIPE, creationflags=0x08000000)
    

    def get_simulated_output(self,reach_ids,constituent,time_step = 'YE'):
        sim = self.model.hbns.get_reach_constituent(constituent,reach_ids,time_step)
        sim.name = 'simulated'
        return sim
    
    def get_observed_data(self,station_ids,constituent,time_step = 'YE'):
        obs = self.dm._get_data(station_ids,constituent,agg_period = time_step).sort_index(level = 'index')
        obs.name = 'observed'
        return obs
     
    
    def compare_simulated_observed(self,station_ids,reach_ids,constituent,time_step,flow_station_ids = None, dropna = False):
        obs = self.get_observed_data(station_ids,constituent,time_step)
        sim = self.get_simulated_output(reach_ids,constituent,time_step = time_step)
        
        #Joing observed and simulated 
        df = sim.join(obs,how = 'outer')
        df = df.loc[(df.index >= obs.index.min()) & (df.index <= obs.index.max())]
        if dropna: df = df.dropna()
        df.columns = ['simulated','observed']
        
        if flow_station_ids is None: flow_station_ids = station_ids
        
        
        # matching flow data
        sim_flow = self.get_simulated_output(reach_ids,'Q',time_step)
        sim_flow.name = 'simulated_flow'
        df = df.join(sim_flow,how = 'inner')
        obs_flow = self.get_observed_data(flow_station_ids,'Q',time_step)
        obs_flow.name = 'observed_flow'
        df = df.join(obs_flow,how='left')
        df.columns = ['simulated','observed','simulated_flow','observed_flow']
        
        # Add metadata
        df.attrs['station_ids'] = station_ids
        df.attrs['reach_ids'] = reach_ids
        df.attrs['constituent'] = constituent
        df.attrs['unit'] = obs.attrs['unit']
        df.attrs['time_step'] = time_step
        df.attrs['flow_station_ids'] = flow_station_ids
        return df
    
    
    
    
    def compare_wplmn(self,station_ids,reach_ids,constituent,unit,flow_station_ids = None,sample=True):
        obs = pd.concat([self.dm.get_wplmn_data(station_id,constituent,unit,'D',samples_only=sample) for station_id in station_ids])
        #sim = self.get_simulated_output(reach_ids,constituent,time_step = 'D',unit = unit)


        if (constituent == 'TSS') & (unit == 'lb'): #convert TSS from lbs to us tons
                obs.loc[:,'value'] = obs.loc[:,'value']/2000
        
        sim = self.model.hbns.get_rchres_data(constituent,reach_ids, unit,'D')

        df = sim.join(obs,how = 'outer')
        df.columns = ['simulated','observed']

        


        
        sim_flow = self.get_simulated_output(reach_ids,'Q','D')
        sim_flow.name = 'simulated_flow'
        df = df.join(sim_flow,how = 'inner')
        if flow_station_ids is None:
            # If wplmn station has flow data use it otherwise use the specifiec wiski station ids
        # matching flow data
            obs_flow = pd.concat([self.dm.get_wplmn_data(station_id,'Q',unit,'D',samples_only=sample) for station_id in station_ids])
        else:
            obs_flow = self.get_observed_data(flow_station_ids,'Q','D')
        
        obs_flow.name = 'observed_flow'
        df = df.join(obs_flow,how='left')
        df.columns = ['simulated','observed','simulated_flow','observed_flow']
        # sim_flow = self.model.hbns.get_rchres_data('Q',reach_ids, 'cfs','D')
        # sim_flow.name = 'simulated_flow'
        # df = df.join(sim_flow,how = 'inner')
        # obs_flow = pd.concat([self.dm.get_data(station_id,'Q','cfs','D') for station_id in station_ids])
        # obs_flow.name = 'observed_flow'
        # df = df.join(obs_flow,how='left')
        # df.columns = ['simulated','observed','simulated_flow','observed_flow']
        
        df.attrs['station_ids'] = station_ids
        df.attrs['reach_ids'] = reach_ids
        df.attrs['constituent'] = constituent
        df.attrs['unit'] = obs.attrs['unit']
        df.attrs['time_step'] = 'D'
        df.attrs['flow_station_ids'] = station_ids
        return df



# Objective Functions sort of

    def aggregate(self,station_ids,reach_ids,constituent,units,time_step,agg_func = 'mean'):
        df = self.compare_simulated_observed(station_ids, reach_ids, constituent, units, time_step)
        
        period = 'M'
        
        if period == 'M':
            grouper = df.index.month
        if period == 'Y':
            grouper = df.index.year
        if period == 'W':
            grouper = df.index.week
        if period == 'D':
            grouper = df.index.dayofyear
        
        df_agg = pd.DataFrame(np.ones((12,3))*np.nan,index = range(1,13),columns = ['simulated','observed','ratio'])
        df_agg.index.name = 'month'
        df = df.groupby(grouper).agg(agg_func)[['simulated','observed']]
        df.columns = ['simulated','observed']
        df['ratio'] = df['observed']/df['simulated']
        df_agg.loc[df.index,df.columns] = df.values
        
        df_agg.loc['Mean'] = df_agg.agg('mean')
        df_agg['ratio'] = df_agg['observed']/df_agg['simulated']

        return df_agg

    def landcover(self,constituent):
        perlnd_names = self.model.uci.table('PERLND','GEN-INFO')['LSID']
        df = self.model.hbns.get_perlnd_data(constituent)*2000 #tons/ac/yr to lbs/acr/year
        df = df[(df.index >= self.start_date) & (df.index <= self.end_date)]
        
        # if self.model.uci.LSID_flag == 1:
        #     print('LSID Error')
        #     return
        
        perland_dict = self.model.uci.opnid_dict['PERLND']
        
        
        dom_lc = int( self.targets['lc_number'][ self.targets['dom_lc'] == 1] )

        df_mean_norm = normalize_mean(df,perland_dict,dom_lc)
        df_mean_norm.loc[df_mean_norm.mean_norm ==0,'mean_norm'] = np.nan
        targets = self.targets.set_index('lc_number')[constituent]
        targets = targets/targets.loc[dom_lc]
        
        df_mean_norm['target'] = targets.loc[df_mean_norm['landcover']].values/df_mean_norm['mean_norm']
        df_mean_norm = df_mean_norm.replace(np.nan,1) #don't make any changes to 0 concentration perlands
        df_mean_norm['uci_name'] = perlnd_names.loc[df_mean_norm.index]
        return df_mean_norm
    
    
    def scour(self,target):     
        # Erosivity adjustment only
        assert((target > 0) & (target < 1))
        scour  = self.model.reports.scour_report()
        #TODO: add check for this
        # Assume all nonpoint values are greater than 0...
        # if depscour is greater than 0
        target = scour['nonpoint']*(1-target)/target # Assuming nonpoint load is set
        adjustment = np.abs(scour['depscour'])/target
        adjustment[(adjustment < 1.05) & (adjustment > .95)] = 1 # Don't change reaches where the depscour is close to the target
        adjustment[adjustment > 1.05] = .95 # Since depscour is negative we have to swap this. I think if I do target/depscour this line would be less confusing
        adjustment[adjustment < .95] = 1.05
        adjustment[scour['depscour'] > 0] = 2 # Double any values where the depscour is positive
        adjustment[scour['LKFG'] == 1] = 1  # Ignore lake flags
        adjustment[np.isnan(adjustment)] = 1
        
        return adjustment

    
    # Methods for updating table values
    def update_table(self,value,operation,table_name,table_id,opnids = None,columns = None,operator = '*',axis = 0):
        self.uci.update_table(value,operation,table_name,table_id,opnids,columns,operator,axis)

    def get_adjustments(self,station_ids,reach_ids, constituent,method):
        
        if method == 'landcover':
            adjustment = self.landcover(constituent)['target'] #pandas series #by opnid
            axis = 0
        elif method =='load': #monthly
            adjustment = self.aggregate(station_ids,reach_ids, constituent,'lb','D',agg_func = 'sum')['ratio'].iloc[0:12] #pandas series by column
            axis = 1
        elif method == 'conc':
            adjustment = self.aggregate(station_ids,reach_ids, constituent,'mg/l','D')['ratio'].iloc[0:12] #pandas series by column
            axis = 1

        else:
            print('Not a valid method')
        return adjustment,axis #ouput dataframe with index representing the opnid


    def update_kser(self,station_ids,reach_ids, constituent,method,opnids = None):   
        #TODO account for the additional comment column
        assert method in ['load','landcover']
        
        adjustment,axis = self.get_adjustments(station_ids,reach_ids, constituent,method)
        if (method == 'landcover') & (opnids is None):
            opnids = adjustment.index.intersection(self.uci.table('PERLND','SED-PARM3',0).index)
                
        self.update_table(adjustment,'PERLND','SED-PARM3', 0,opnids,['KSER'], '*',axis)
 
    
    def update_qualprop(self,station_ids,reach_ids, constituent,method,table_name, opnids = None, months = [1,2,3,4,5,6,7,8,9,10,11,12],threshold = 0, max_change = 1000, update_alg = '*'):
        assert method in ['conc','load','landcover']
        assert table_name in ['MON-IFLW-CONC','MON-GRND-CONC']
        assert max_change >= threshold
        
       
        
        threshold = threshold/100
        max_change = max_change/100
        
        table_id = {'N':1,
                    'TKN':0,
                    'OP':2,
                    'TP':3}[constituent]
                    #nutrient_id-1 # Shift as tables in the uci dictionary are stored starting at the 0 index
        
        adjustment,axis = self.get_adjustments(station_ids,reach_ids, constituent,method)
        if opnids is None:
            opnids = adjustment.index.intersection(self.uci.table('PERLND',table_name,table_id).index)
        
        if method == 'load':
            adjustment = adjustment.loc[months]
        
        if method == 'landcover':
            adjustment = adjustment.loc[opnids]
  
    
        # Apply threshold adjustment
        adjustment[np.isnan(adjustment)] = 1
        adjustment[np.abs((adjustment-1)) <= threshold] = 1 # don't change values below threshold
        direction = np.sign(adjustment - 1)[np.abs(adjustment - 1) > max_change]
        adjustment[np.abs(adjustment-1) > max_change] = 1+ direction*max_change # Formla assumes it is flux/model
        #Note that in uci.update_table() there is further screening to account for adjustments below the model precision
        
        #column_prefix = self.uci.get_table_info('PERLND',table_name)['NAME'].str[:3].iloc[1]
        months = np.array(months)
        column_prefix = self.uci.table('PERLND',table_name,table_id).columns.str[:3][0]
        columns = [column_prefix+helpers.get_months(month) for month in months]
        
        self.uci.update_table(adjustment.values,'PERLND',table_name,table_id,opnids,columns,update_alg,axis)
        
    def update_erosivity(self, opnids = None):
        adjustment,axis = self.get_adjustments(0,'scour')
        
        table = self.uci.table('RCHRES','SILT-CLAY-PM',0)
        
        if opnids is None:
            opnids = adjustment.index.intersection(table.index)
        
        self.update_table(adjustment,'RCHRES','SILT-CLAY-PM', 0,opnids,'M','*',axis)
        self.update_table(adjustment,'RCHRES','SILT-CLAY-PM', 1,opnids,'M','*',axis)

    
    def save_output(self, constituent,station_ids,reach_ids,time_step = 'D',flow_station_ids=None,drng_area = None,save_path = None,start_year = 1996, end_year = 2100):
        if save_path is None:
            save_path = self.output_path
        if drng_area is None:
            drng_area = self.uci.network.drainage_area(reach_ids)
            #drng_area = sum([self.uci.network.drainage_area(reach_id) for reach_id in reach_ids])        
        
        run_number = self.run
        
        if flow_station_ids is None:
            flow_station_ids = station_ids
            
        if constituent == 'Q':
            constituent = 'Q'
            units = 'cfs'
            df = self.compare_simulated_observed(flow_station_ids,reach_ids,constituent,
                                                time_step = time_step)
    
    
    
            #TODO: Move the exporting of these tables into the report module to decouple things
            metrics.hydro_stats(df.dropna(),drng_area).to_csv(save_path.joinpath(f'hydrostats_{station_ids}'))
            self.model.reports.annual_water_budget('PERLND').to_csv(save_path.joinpath(f'{station_ids}_annual_perlnd_water_budget.csv'))
            self.model.reports.annual_water_budget('IMPLND').to_csv(save_path.joinpath(f'{station_ids}_annual_implnd_water_budget.csv'))
            self.model.reports.annual_water_budget('RCHRES').to_csv(save_path.joinpath(f'{station_ids}_annual_perlnd_rchres_budget.csv'))
            self.model.reports.ann_avg_watershed_loading(constituent,reach_ids).to_csv(save_path.joinpath('annual_runoff.csv'))
            #self.model.reports.monthly_runoff().to_csv(save_path.joinpath('monthly_runoff.csv'))

            
           #self.model.reports.avg_ann_yield('Q',reach_ids).to_csv(save_path.joinpath(f'{station_ids}_avg_ann_outflows.csv'))
            #self.model.reports.avg_monthly_outflow().to_csv(save_path.joinpath(f'{station_ids}_avg_monthly_outflows.csv'))
            
            figures.contTimeseries(df,station_ids,constituent,units,save_path)
            figures.FDCexceed(df.dropna(),station_ids,constituent,units,save_path)
            figures.scatter(df.dropna(),station_ids,constituent,units,save_path)
            figures.monthly_bar(metrics.monthly(df.dropna(),units),station_ids,constituent,units,save_path)
        
        else:
            assert constituent in ['TSS','TP','OP','N','TKN']
            df= self.compare_simulated_observed(
                station_ids,
                reach_ids,
                constituent,
                time_step,
                flow_station_ids = flow_station_ids)
            
            if df.empty:
                print(f'No Observation Data found for {constituent} at {station_ids}')
                return
            reachid_str = '_'.join([str(reach_id) for reach_id in reach_ids])

            figures.timeseries(df,station_ids,constituent,'mg/l',save_path = save_path.joinpath(f'{constituent}_Timeseries_{reachid_str}_{run_number}.png'))
            
            
            df =df.loc[(df.index.year >= start_year) & (df.index.year <= end_year)]

            stats = metrics.stats(df.dropna(subset=['observed']),'mg/l')[['observed','simulated','per_error','abs_error']]
            stats.to_csv(save_path.joinpath(f'{constituent}_stats_{reachid_str}_{run_number}.csv'))
        
            figures.scatter(df.dropna(subset=['observed']),station_ids,constituent,'mg/l',save_path.joinpath(f'{constituent}_Scatter_{reachid_str}_{run_number}.png'))
            figures.FDCexceed(df.dropna(subset=['observed']),station_ids,constituent,'mg/l',save_path.joinpath(f'{constituent}_Exceedence_{reachid_str}_{run_number}.png')) 
            figures.timeseries(df,station_ids,constituent,'mg/l',save_path.joinpath(f'{constituent}_Timeseries_{reachid_str}_{run_number}.png'))
            
            if len(df.dropna(subset = ['observed','observed_flow'])) > 10:    
                figures.rating(df.dropna(subset = ['observed','observed_flow']),station_ids,constituent,'mg/l',save_path.joinpath(f'{constituent}_Rating_{reachid_str}_{run_number}.png'))
                figures.LDC(df.dropna(subset = ['observed','observed_flow']),station_ids,constituent,'mg/l',time_step = time_step, save_path = save_path.joinpath(f'{constituent}_LDC_{reachid_str}_{run_number}.png'))

            
            

def chuck(adjustment,table):
    # If increasing monthly concentration increase the minimum concnetration value of Mi and Mi+1
    # If decreasing monthly concentration decrease the maximum concnetration value of Mi and Mi+1
    # If concnetration values are equal increase both equally
    table['dummy'] = table.iloc[:,0]
    zero_table = table.copy()*0
    count_table = zero_table.copy()
    for index, value in enumerate(adjustment[0]):
            next_index = index+1             
            if value > 1:
                for row,(a,b) in enumerate(zip(table.iloc[:,index].values, table.iloc[:,next_index].values)):
                    zero_table.iloc[row,index+np.nanargmin([a,b])] += np.nanmin([a,b])*value
                    count_table.iloc[row,index+np.nanargmin([a,b])] += 1
            elif value < 1:
                for row,(a,b) in enumerate(zip(table.iloc[:,index].values, table.iloc[:,next_index].values)):
                    zero_table.iloc[row,index+np.nanargmax([a,b])] += np.nanmax([a,b])*value
                    count_table.iloc[row,index+np.nanargmax([a,b])] += 1
    
    
    zero_table.drop('dummy',axis=1,inplace=True)
    count_table.drop('dummy',axis=1,inplace=True)
    
    zero_table[count_table == 0] = table[count_table==0]
    count_table[count_table == 0] = 1
    zero_table = zero_table/count_table
    return zero_table       

def normalize_mean(df,perland_dict,dom_lc):
     df_mean = df.mean().rename('mean').to_frame()
     df_mean['mean_norm'] = df_mean['mean'].values
     
     perlands = df_mean.index
     perland_dict['landcover_group'] = (perland_dict['landcover'] == perland_dict['landcover'].iloc[0]).cumsum()
     
     metzones = np.array([perland_dict.loc[perland,'metzone'] for perland in perlands])
     landcover = np.array([perland_dict.loc[perland,'landcover'] for perland in perlands])
     landcover_groups = np.array([perland_dict.loc[perland,'landcover_group'] for perland in perlands])

     for group in np.unique(landcover_groups):
         dom_perlnd = perland_dict.index[(perland_dict['landcover_group'] == group) & (perland_dict['landcover'] == dom_lc)]
         df_mean.loc[landcover_groups == group,'mean_norm'] = df_mean.loc[landcover_groups == group,'mean_norm']/df_mean.loc[dom_perlnd,'mean_norm'].values
         
     df_mean['metzones'] = metzones
     df_mean['landcover'] = landcover
     df_mean['perlands'] = perlands
     df_mean.sort_values(by = ['metzones','perlands'],inplace = True)
     return df_mean


# def column_prefix():
#     months = np.array(months)
#     column_prefix = self.uci.table('PERLND',table_name,table_id).columns.str[:3][0]
#     columns = [column_prefix+ch.get_months(month) for month in months]

def threshold(adjustment,threshold,max_change):
    # Apply threshold adjustment
    adjustment[np.isnan(adjustment)] = 1
    adjustment[np.abs((adjustment-1)) <= threshold] = 1 # don't change values below threshold
    direction = np.sign(adjustment - 1)[np.abs(adjustment - 1) > max_change]
    adjustment[np.abs(adjustment-1) > max_change] = 1+ direction*max_change # Formla assumes it is flux/model
    #Note that in uci.update_table() there is further screening to account for adjustments below the model precision
    return adjustment  


#class hydrologyCalibrator(calibrator):
    
#class nutrientCalibrator(calibrator):
        
class sedimentCalibrator(calibrator):
    
    def update_kser(self,method,opnid = None):   
        #TODO account for the additional comment column
        assert method in ['load','landcover','sftl']
        
        table = self.uci.table('PERLND','SED-PARM3',0,False)
        
        
        if method == 'load':
            adjustment = self.compare(0,aggregate = True).loc['Mean']['ratio']      
        elif method == 'landcover':
            adjustment = self.landcover(0)['target']
            table = self.uci.table('PERLND','SED-PARM3',0)
            if opnid == None:
                opnid = table.index
            adjustment = np.array(adjustment.loc[opnid])[:,None]
        elif method == 'sftl':
            adjustment = self.sftl()
        
        self.uci.replace_table('PERLND','SED-PARM3',0)
    
    def update_erosivity(self,param = 'M',opnid = None,update_alg = '*'):
        adjustment = self.scour()  
        table = self.uci.table('RCHRES','SILT-CLAY-PM',0)
        if opnid == None:
            opnid = table.index
        adjustment = np.array(adjustment.loc[opnid])[:,None]
        self.uci.update_table(adjustment,'RCHRES','SILT-CLAY-PM',table_id = 0,opnid = opnid,columns = [param],update_alg = update_alg)
        
        adjustment = self.scour()
        adjustment = np.array(adjustment.loc[opnid])[:,None]
        self.uci.update_table(adjustment,'RCHRES','SILT-CLAY-PM',table_id = 1,opnid = opnid,columns = [param],update_alg = update_alg) 
    
        
    def fit_param(self,param,m_factor,N = 2,opnid = None,run = None):
        bounds = {'M':[.000000001,.01,2,5], #maxlow,low,high,maxhigh
                  'TAUCD':[.001,.01,.3,1],
                  'TAUCS':[.01,.05,.5,3]}
        if run == None:
            run = self.run
        
        data = self.load_data('scour',N=10000)
        data = data.loc[:,range(run-N+1,run+1),:]
    
        if opnid == None:
            opnid = data.reset_index(level=[1]).index.unique() # assumes multiindex
        
        for index in opnid:
            if any(data.loc[index]['LKFG'] == 0):
                x = data.loc[index]['depscour']
                y = data.loc[index][param]
                linear_model=np.polyfit(x,y,1)
                linear_model_fn=np.poly1d(linear_model)
                m = linear_model_fn(-data.loc[index]['nonpoint'].iloc[1]*.25)
                if m < bounds[param][0]:
                    m = bounds[param][0]
                if m > bounds[param][3]:
                    m = bounds[param][3]
                self.update_table('RCHRES','SILT-CLAY-PM',0,m,'set',opnid = index,columns = [param]) #mod.update_table(operation,table_name,table_id,adjustment,operator,opnids,columns)
                self.update_table('RCHRES','SILT-CLAY-PM',1,m*m_factor,'set',opnid = index,columns = [param]) #mod.update_table(operation,table_name,table_id,adjustment,operator,opnids,columns)
    
    def erosivity(self,m_factor,param = 'M',opnid = None,run = None,iterations = 1):
        
        if run == None:
            run = self.run
        
        # run model updating erosivity for N iterations
        for iteration in range(iterations):
            self.update_erosivity(param = param,opnid = opnid)
            self.run_model() # creates the run+1 uci file and runs it using WinHspfLT
            run = run + 1
            self.load_model(run)
            self.save_data()
     
    
        self.fit_param(param,m_factor,iterations+1,opnid,run)
        self.run_model() # creates the run+1 uci file and runs it using WinHspfLT
        
        run = run + 1
        self.load_model(run)
        self.save_data()

    def scour(hbn,uci):     
        # Erosivity adjustment only
        scour  = reports.scour_report(hbn,uci)
        #TODO: add check for this
        # Assume all nonpoint values are greater than 0...
        # if depscour is greater than 0
        target = scour['nonpoint']*.25 # Assuming nonpoint load is set
        adjustment = np.abs(scour['depscour'])/target
        adjustment[(adjustment < 1.05) & (adjustment > .95)] = 1 # Don't change reaches where the depscour is close to the target
        adjustment[adjustment > 1.05] = .95 # Since depscour is negative we have to swap this. I think if I do target/depscour this line would be less confusing
        adjustment[adjustment < .95] = 1.05
        adjustment[scour['depscour'] > 0] = 2 # Double any values where the depscour is positive
        adjustment[scour['LKFG'] == 1] = 1  # Ignore lake flags
        adjustment[np.isnan(adjustment)] = 1
        
        return adjustment


    



# -*- coding: utf-8 -*-
"""
Created on Wed May  8 09:40:51 2024

@author: cfreema
"""

#%% Add libraries

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from datetime import datetime
from scipy.stats import pearsonr
from pathlib import Path


#%% Minnesota Color Scheme and Font

# Minnesota Brand Colors
COLORS = {'Minnesota Blue': '#003865',
          'Minnesota Green': '#78BE21',
          'White': '#FFFFFF',
          'Black': '000000',
          'Accent Teal': '#008EAA',
          'Accent Green': '#0D5257',
          'Accent Orange': '#8D3F2B',
          'Accent Purple': '#5D295F',
          'Extended Accent Blue Gray': '#A4BCC2',
          'Extended Accent Cream': '#F5E1A4',
          'Extended Accent Sky Blue': '#9BCBEB',
          'Extended Accent Gold': '#FFC845',
          'Dark Gray': '#53565A',
          'Medium Gray': '#97999B',
          'Light Gray': '#D9D9D6',
          'Orange': '#E57200'}

# Minnesota Substitute Typography
CFONT = {'fontname': 'Calibri'}

#%% Monthly Bar Plot
def monthly_bar(df, station_id, constituent, units, save_path=None, yaxis=None, ylimit=None):
    '''
    Returns a timeseries plot of hourly simulated concentrations and observed concentrations 
    with the same date range as the input data (df). The hourly simulated concentrations are 
    represented with a line and the observed concentrations are represented with points.

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 2 columns: simulated, observed.
        -The dates must be in the datetime Python format.
        -The date range must already be filtered to the date range you wish to plot in the 
         figure and should be on an hourly timestep.
        -The hourly simulated concentrations must be paired with observed concentrations, where 
         grab samples exist.
        -If you wish to plot concentrations from more than one station, there should still only be one 
         'obs' column, i.e., no stations overlapping in dates.
        -In pre-processing, the time of the observed samples may need to be adjusted to the nearest 
         top of the hour in order to merge the observed data with the simulated data.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen) 
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
   
            
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_timeseries_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'
    
    yaxis: Python string
        -Options: None, log, Log
        -Default is None and will create a plot with a linear y-axis.
        -yaxis = log (or Log) will create a plot with a log y-axis.
        -Use log when you have a dataset with a high range of concentrations (i.e. TSS).
        
    ylimit: integer
        -Insert a value that will be the upper limit of the y-axis (e.g. ylimit = 400)
        -Default is None and will use the dataset to set the y-axis limits.

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    
    units = df.attrs['unit']
    
    assert units in ['mg/l','lb','cfs']
    
    if units in ['cfs','mg/l']:
        df = df[['observed','simulated']].groupby(df.index.month).mean()
    else:
        df = df[['observed','simulated']].groupby(df.index.month).sum()

    
    # cf2cfs = {'hourly':3600,
    #           'daily':86400,
    #           'monthly':2592000,
    #           'yearly':31536000,
    #           'h':3600,
    #           'D':86400,
    #           'M':2592000,
    #           'Y':31536000}
    
    # if units == 'mg/l':
    #     df['simulated'] = 2.20462e-6*df['simulated']/0.0353147*df['simulated_flow']*cf2cfs[time_step]
    #     df['observed'] = 2.20462e-6*df['observed']/0.0353147*df['observed_flow']*cf2cfs[time_step]
    #     units = 'lb'
    
    
    
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # months = {1:'JAN', 2:'FEB', 3:'MAR', 4:'APR', 5:'MAY', 6:'JUN', 7:'JUL',
    #        8:'AUG', 9:'SEP', 10:'OCT', 11:'NOV', 12:'DEC'}
    
    
    
    #index = ['January','Febuary','March','April','May','June','July','August','September','October','November','December']
    df[['observed','simulated']].plot.bar(ax=ax,color = {'simulated':COLORS.get('Orange'),'observed':COLORS.get('Minnesota Blue')})
    stations = ', '.join(station_id)
    plt.title(f'Simulated and Observed {constituent} \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Month', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(['Observed', 'Simulated'], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    
    #Axis settings depending on user input
    if yaxis == 'log' or yaxis == 'Log':
        # Convert y axis to log scale
        plt.yscale('log')
        # Add axis gridlines
        plt.grid(which='major', axis='y', linewidth=0.3)
        plt.grid(which='minor', axis='y', linewidth=0.3)
        
    else:
        # Add y axis gridlines and set the y axis minimum to 0
        ax.set_axisbelow(True)
        plt.grid(which='major', axis='y', linewidth=0.3)
        plt.ylim(bottom=0)
        
    if ylimit is not None:
        # Define the upper limit of the y-axis based on user input
        plt.ylim(top = ylimit)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_monthly_bar' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')
    
    return fig, ax

#%% Timeseries
def timeseries(df, station_id, constituent, units, save_path=None, yaxis=None, ylimit=None):
    '''
    Returns a timeseries plot of hourly simulated concentrations and observed concentrations 
    with the same date range as the input data (df). The hourly simulated concentrations are 
    represented with a line and the observed concentrations are represented with points.

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 2 columns: simulated, observed.
        -The dates must be in the datetime Python format.
        -The date range must already be filtered to the date range you wish to plot in the 
         figure and should be on an hourly timestep.
        -The hourly simulated concentrations must be paired with observed concentrations, where 
         grab samples exist.
        -If you wish to plot concentrations from more than one station, there should still only be one 
         'obs' column, i.e., no stations overlapping in dates.
        -In pre-processing, the time of the observed samples may need to be adjusted to the nearest 
         top of the hour in order to merge the observed data with the simulated data.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen) 
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            mg/L (milligrams per liter)
            ppm (parts per million)
            ug/L (micrograms per liter)
            ppb (parts per billion)
            
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_timeseries_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'
    
    yaxis: Python string
        -Options: None, log, Log
        -Default is None and will create a plot with a linear y-axis.
        -yaxis = log (or Log) will create a plot with a log y-axis.
        -Use log when you have a dataset with a high range of concentrations (i.e. TSS).
        
    ylimit: integer
        -Insert a value that will be the upper limit of the y-axis (e.g. ylimit = 400)
        -Default is None and will use the dataset to set the y-axis limits.

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Plot simulated TSS concentrations (line)
    sim, = ax.plot(df['simulated'], label = 'sim', color = COLORS.get('Orange'), linewidth = '0.9')  
    # Plot observed TSS concentrations (points)
    obs, = ax.plot(df['observed'], '*', mfc = 'none', label = 'obs', mec = COLORS.get('Minnesota Blue'))  
    # Add plot title
    stations = ', '.join(station_id)
    plt.title(f'Simulated and Observed {constituent} Concentrations \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Date', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(handles=[sim, obs], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    
    #Axis settings depending on user input
    if yaxis == 'log' or yaxis == 'Log':
        # Convert y axis to log scale
        plt.yscale('log')
        # Add axis gridlines
        plt.grid(which='major', axis='y', linewidth=0.3)
        plt.grid(which='minor', axis='y', linewidth=0.3)
        
    else:
        # Add y axis gridlines and set the y axis minimum to 0
        plt.grid(which='major', axis='y', linewidth=0.3)
        plt.ylim(bottom=0)
        
    if ylimit is not None:
        # Define the upper limit of the y-axis based on user input
        plt.ylim(top = ylimit)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_timeseries_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')
    
    return fig, ax

#%% Simulated versus Observed Concentrations Plot (Scatter): scatter

def scatter(df, station_id, constituent, units, save_path=None):
    '''
    Returns a scatter plot of simulated and observed concentrations. 
    A 1:1 fit indicates a perfect relationship between simulated and observed concentrations.

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 3 columns: sim, obs.
        -The simulated concentrations must be paired with observed concentrations, where grab samples exist.
        -If you wish to plot concentrations from more than one station, there should still only be one 
         'obs' column, i.e., no stations overlapping in dates.
        -In pre-processing, the time of the observed samples may need to be adjusted to the nearest top of 
         the hour in order to merge the observed data with the simulated data.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            mg/L (milligrams per liter)
            ppm (parts per million)
            ug/L (micrograms per liter)
            ppb (parts per billion)  
    
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_scatter_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    # Calculate correlation coefficient
    corr, _ = pearsonr(df['observed'],df['simulated'])
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Plot observed versus simulated TSS concentrations (points)
    ax.plot(df['observed'], df['simulated'], 'o', alpha = 0.3, color = COLORS.get('Minnesota Green'))     
    # Add the correlation coefficient to the plot
    plt.text(0.1, 0.9, f'r = {corr:.2f}', transform=plt.gca().transAxes)
    # Add plot title
    stations = ', '.join(station_id)
    plt.title(f'Simulated versus Observed {constituent} Concentrations \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel(f'Observed {constituent} ({units})', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'Simulated {constituent} ({units})', **CFONT, fontsize = 10) 
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Add y axis gridlines, set the x and y axis minimums to 0, and make the axes square
    plt.grid(which='major', axis='y', linewidth=0.3)
    plt.axis('square')
    plt.ylim(bottom=0)
    plt.xlim(left=0)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_scatter_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax


#test_df = create_test_data()

#fig, ax = scatter(test_df,station_id, 'TSS', 'mg/L')


#%% Concentration vs Flow (Rating Curve): rating

def rating(df, station_id, constituent, units, save_path=None):
    '''
    Returns a concentration versus flow rating curve for both simulated and observed datasets. 

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 4 columns: simulated, simulated_flow, observed, observed_flow.
        -This function assumes the flow units are in cubic feet per second (cfs).
        -The simulated concentrations & flow must be paired with observed concentrations & flow.
        -If you wish to plot concentrations and flow from more than one station, there should 
         still only be one 'obs conc' column and one 'obs flow (cfs)' column
        -In pre-processing, the time of the observed samples and flow may need to be adjusted 
         to the nearest 15-minutes or nearest top of the hour in order to merge the observed data 
         with the simulated data.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)   
        
    units: Python string
        -Input constituent units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            mg/L (milligrams per liter)
            ppm (parts per million)
            ug/L (micrograms per liter)
            ppb (parts per billion)  
            
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_rating_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'
    
    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Plot observed concentration versus flow (points)
    obs, = ax.plot(df['observed_flow'], df['observed'], '*', label = 'obs', color = COLORS.get('Minnesota Blue'))     
    # Plot simulated concentration versus flow (points)
    sim, = ax.plot(df['simulated_flow'], df['simulated'], 'v', mfc = 'none', label = 'sim', mec = COLORS.get('Orange'))   
    # Add plot title
    stations = ', '.join(station_id)
    plt.title(f'Simulated and Observed {constituent} Concentration vs Flow \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Flow (cfs)', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(handles=[obs, sim], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Convert x and y axis to log scale
    plt.yscale('log')
    plt.xscale('log')
    # Add axis gridlines
    plt.grid(which='major', axis='both', linewidth=0.3)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_rating_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax


#test_df = create_test_data()

#fig, ax = rating(test_df,station_id, 'PO4', 'ug/L')

#%% Load Duration Curve: LDC

# # Create random data for input into LDC function
# def create_test_data():
#     start_date = datetime(2020,5,20)
#     end_date = datetime(2023,9,30)
#     date_range = ((end_date - start_date).days + 1)

#     random_days = np.random.randint(date_range, size=45)
#     obs_dates = pd.to_datetime(start_date) + pd.to_timedelta(random_days, unit='days')

#     test_df_obs = pd.DataFrame({'date' : obs_dates,
#                                 'obs load': np.random.randint(9000, size = 45).astype(float),
#                                 'obs flow': np.random.randint(1000, size = 45).astype(float)})

#     test_df_obs['obs rank'] = test_df_obs['obs flow'].rank(method = 'first', ascending = False)
#     test_df_obs['obs exceed %'] = test_df_obs['obs rank'] / (len(test_df_obs) + 1) * 100

#     sim_dates = pd.date_range(start='1/1/2020',end='12/31/2023')

#     test_df_sim = pd.DataFrame({'date' : sim_dates,
#                                 'sim load': np.random.randint(9000, size = 1461).astype(float),
#                                 'sim flow': np.random.randint(1000, size = 1461).astype(float)})

#     test_df_sim['sim rank'] = test_df_sim['sim flow'].rank(method = 'first', ascending = False)
#     test_df_sim['sim exceed %'] = test_df_sim['sim rank'] / (len(test_df_sim) + 1) * 100

#     test_df = test_df_sim.merge(test_df_obs, on = 'date', how = 'left')
    
#     return test_df


def LDC(df, station_id, constituent, units, time_step, save_path=None):
    '''
    Returns a load duration curve - constituent load versus flow exceedance percentage 
    for both simulated and observed datasets. 

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 5 columns: date, sim load, sim exceed %, obs load, obs exceed %.
        -Sim exceed % and obs exceed % represent flow exceedance (not load exceedance).
        -The entire modeled period of record may be used for the simulated data.
        -If you wish to use load and flow observational data from more than one station, there should 
          still only be one 'obs load' column and one 'obs exceed %' column.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)  
            
    units: Python string
        -Input constituent units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            tons per day (tons/day)
            pounds per day (lb/day)

    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_LDC_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    assert units in ['mg/l','lb']
    
    cf2cfs = {'hourly':3600,
              'daily':86400,
              'monthly':2592000,
              'yearly':31536000,
              'h':3600,
              'D':86400,
              'M':2592000,
              'Y':31536000}
    
    if units == 'mg/l':
        df['simulated'] = 2.20462e-6*df['simulated']/0.0353147*df['simulated_flow']*cf2cfs[time_step]
        df['observed'] = 2.20462e-6*df['observed']/0.0353147*df['observed_flow']*cf2cfs[time_step]
        units = 'lb'
        
    df['simulated_flow rank'] = df['simulated_flow'].rank(method = 'average', ascending = False)
    df['simulated_flow exceed %'] = df['simulated_flow rank'] / (len(df) + 1) * 100
    df['observed_flow rank'] = df['observed_flow'].rank(method = 'average', ascending = False)
    df['observed_flow exceed %'] = df['observed_flow rank'] / (len(df) + 1) * 100

    
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)   
    # Plot simulated load versus flow exceedance (points)
    sim, = ax.plot(df['simulated_flow exceed %'], df['simulated'], 'v', mfc = 'none', label = 'sim', mec = COLORS.get('Orange'))
    # Plot observed load versus flow exceedance (points)
    obs, = ax.plot(df['observed_flow exceed %'], df['observed'], '*', label = 'obs', color = COLORS.get('Minnesota Blue'))  
    # Add plot title
    stations = ', '.join(station_id)
    plt.title(f'{constituent} Load Duration Curve \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Flow Exceedance (%)', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(handles=[obs, sim], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Convert y axis to log scale
    plt.yscale('log')
    # Add axis gridlines
    plt.grid(which='major', axis='both', linewidth=0.3)
    plt.grid(which='minor', axis='y', linewidth=0.3)
    
    #Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_LDC_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax

# #test_df = create_test_data()

# #fig, ax = LDC(test_df,station_id, 'OP', 'lbs/day')

#%% Timeseries Plot: contTimeseries

def contTimeseries(df, station_id, constituent, units, save_path=None):
    '''
    Returns a timeseries plot of simulated and observed continuous data with the same date range as
    the input data (df).

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 2 columns: observed, simulated
        -The dates must be in the datetime Python format.
        -The date range must already be filtered to the date range and timestep you wish to plot 
         in the figure.
        -Each simulated data point must have a corresponding observed data point. 
        -For TSS, N+N, PO4, TP, and TKN, the simulated data should be from HSPF and the observed data
         should be from FLUX (WPLMN).
        -The suggested units for each of the constituents are as follows:
                -flow: cubic feet per second (cfs)
                -temp: degrees Fahrenheit (ºF)
                -DO: milligrams per liter (mg/L)
                -TSS: kilograms per day (kg/d)
                -N+N: kilograms per day (kg/d)
                -PO4: kilograms per day (kg/d)
                -TP: kilograms per day (kg/d)
                -TKN: kilograms per day (kg/d)
        -If you wish to plot observed data from more than one station, there should still only be one 
         column for obs, i.e., no observed data overlapping in dates.
        
    station_id : Python list
        -Include one or more station ID numbers (i.e., ['H49009001', 'H25019002']).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis        
        -Options include: 
            Flow
            Temp (water temperature)
            DO (dissolved oxygen)
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            cfs
            ºF (degrees F)
            ºC (degrees C)
            mg/L (milligrams per liter)
            ppm (parts per million)
            kg/d (kilograms per day)

    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_contTimeseries_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'
        
    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Plot observed flow
    obs, = ax.plot(df['observed'], label = 'observed', color = COLORS.get('Minnesota Blue'), linewidth = '0.9')  
    # Plot simulated flow
    sim, = ax.plot(df['simulated'], label = 'simulated', color = COLORS.get('Orange'), linewidth = '0.5')  
    # Add plot title
    stations = ', '.join(station_id)
    plt.title(f'Simulated and Observed {constituent} \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Date', **CFONT, fontsize = 10)   
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(handles=[obs, sim], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Add y axis gridlines and set the y axis minimum to 0
    plt.grid(which='major', axis='y', linewidth=0.3)
    plt.ylim(bottom=0)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_contTimeseries_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax

#test_df = create_test_data()

#fig, ax = contTimeseries(test_df,station_id, 'Nitrate + Nitrite', 'kg/d')

#%% Duration Curve/Exceedance Plot: FDCexceed

# Create random data for input into FDCexceed function
def create_test_data():
    start_date = datetime(2020,5,20,8,0,0)
    end_date = datetime(2023,9,30,17,0,0)
    date_range = ((end_date - start_date).days + 1) * 24

    random_days = np.random.randint(date_range, size=80)
    dates = pd.to_datetime(start_date) + pd.to_timedelta(random_days, unit='hours')

    test_df_sim = pd.DataFrame({'date' : dates,
                                'sim': np.random.randint(500,size = 80).astype(float)})

    test_df_sim['simulated rank'] = test_df_sim['simulated'].rank(method = 'first', ascending = False)
    test_df_sim['simulated exceed %'] = test_df_sim['simulated rank'] / (len(test_df_sim) + 1) * 100

    test_df_obs = pd.DataFrame({'date' : dates,
                                'obs': np.random.randint(500,size = 80).astype(float)})

    test_df_obs['obs rank'] = test_df_obs['observed'].rank(method = 'first', ascending = False)
    test_df_obs['obs exceed %'] = test_df_obs['observed rank'] / (len(test_df_sim) + 1) * 100

    test_df = test_df_obs.merge(test_df_sim,on='date')
    
    return test_df

def FDCexceed(df, station_id, constituent, units, save_path=None):
    '''
    Returns a flow duration curve or exceedance plot for simulated and observed data.

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 2 columns: simulated, observed
        -In data preprocessing, the simulated data should be paired with observed data.
        -If you wish to use observed data from more than one station, there should still only be one 
         'obs' column, i.e., no overlapping observed data.
        -Exceedance percentage should be calculated by ranking data (highest is ranked #1), 
         then calculate exceedance percentage as follows:
            P = (m/n+1) * 100, where:
                P = exceedance percentage
                m = the ranking of all data for the period of record
                n = the total number of data points
                
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            Flow
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            cfs
            mg/L (milligrams per liter)
            ppm (parts per million)
            ug/L (micrograms per liter)
            ppb (parts per billion)
            
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_FDCexceed_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    
    
    df['simulated rank'] = df['simulated'].rank(method = 'average', ascending = False)
    df['simulated exceed %'] = df['simulated rank'] / (len(df) + 1) * 100
    df['observed rank'] = df['observed'].rank(method = 'average', ascending = False)
    df['observed exceed %'] = df['observed rank'] / (len(df) + 1) * 100

    
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Sort obs in descending order, then plot observed FDC/exceedance plot
    df = df.sort_values(by = 'observed', ascending = False)
    obs, = ax.plot(df['observed exceed %'], df['observed'], label = 'observed', color = COLORS.get('Minnesota Blue'))  
    # Sort sim in descending order, then plot simulated FDC/exceedance plot
    df = df.sort_values(by = 'simulated', ascending = False)
    sim, = ax.plot(df['simulated exceed %'], df['simulated'], label = 'simulated', color = COLORS.get('Orange'))  
    # Add plot title
    stations = ', '.join(station_id)
    if constituent == 'Flow' or constituent == 'flow':
        plt.title('Flow Duration Curves \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    else: 
        plt.title(f'{constituent} Exceedance Plot \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    plt.xlabel('Exceedance Percentage', **CFONT, fontsize = 10) 
    # Add y label
    plt.ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    plt.legend(handles=[obs, sim], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Add axis gridlines
    plt.grid(which='major', axis='both', linewidth=0.3)
    plt.grid(which='minor', axis='y', linewidth=0.3)
    # Convert y axis to log scale
    plt.yscale('log')
    # Set the x axis minimum to 0
    plt.xlim(left=0)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_FDCexceed_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax



#test_df = create_test_data()

#fig, ax = FDCexceed(test_df,station_id, 'flow', 'cfs')

#%% Flow Duration Curve or Exceedance Plot with Points: FDCexceedPoints

def FDCexceedPoints(df, station_id, constituent, units, save_path=None):
    '''
    Returns a duration curve (flow) or exceedance plot (sediment and nutrients)
    for observed data with points representing paired (not ranked) simulated data.

    Parameters
    ----------
    df : Pandas Dataframe
        -Dataframe with 2 columns: simulated, observed
        -In data preprocessing, the simulated data should be paired with observed data.
        -If you wish to use observed data from more than one station, there should still only be one 
         'obs' column, i.e., no overlapping observed data.
        -Exceedance percentage should be calculated by ranking data (highest is ranked #1), 
         then calculate exceedance percentage as follows:
            P = (m/n+1) * 100, where:
                P = exceedance percentage
                m = the ranking of all data for the period of record
                n = the total number of data points
                
    station_id : Python list
        -Include one or more station ID numbers (i.e., H49009001).
        -This argument is only used to include the station ID(s) in the plot title.
        -Example: ['H49009001'] or ['W49012002', 'W49012003']
        
    constituent: Python string
        -Input constituent as you'd like it to appear on the figure title and axis
        -Options include: 
            Flow
            TSS (total suspended solids)
            N+N (nitrate plus nitrite nitrogen)
            PO4 (orthophosphate phosphorus)
            TP (total phosphorus)
            TKN (total Kjeldahl nitrogen)
            
    units: Python string
        -Input units as you'd like it to appear on the figure axis (without parentheses)
        -Make sure the units you enter match the units of the data in your input df
        -Options include:
            cfs
            mg/L (milligrams per liter)
            ppm (parts per million)
            ug/L (micrograms per liter)
            ppb (parts per billion)
            
    save_path: Python string
        -Default is None and will not save the figure anywhere.
        -Enter the folder path OR the file path (it is recommended to save the file as a .png).
        -If a folder path is entered, the file name will default to 'constituent_FDCexceed_stationID(s)'
        -Make sure to use forward slashes (/) instead of backslashes (\).
        -File path example: 'C:/Users/cfreema/Desktop/delete/figure_6.png'
        -Folder path example: 'C:/Users/cfreema/Desktop/delete/'

    Returns
    -------
    fig : Figure object of matplotlib.figure module
    ax : AxesSubplot object of matplotlib.axes._subplots module

    '''
    
    df['observed rank'] = df['observed'].rank(method = 'first', ascending = False)
    df['observed exceed %'] = df['observed rank'] / (len(df) + 1) * 100

    
    # Create a figure containing a single axes
    fig, ax = plt.subplots(figsize = (7,4), dpi = 600)
    # Sort obs in descending order, then plot observed FDC/exceedance plot and paired sim conc
    df = df.sort_values(by = 'observed', ascending = False)
    sim, = ax.plot(df['observed exceed %'], df['simulated'], '.', label = 'simulated', color = COLORS.get('Orange'))  
    obs, = ax.plot(df['observed exceed %'], df['observed'], label = 'observed', color = COLORS.get('Minnesota Blue'))  
# START HERE 
    # Add plot title
    stations = ', '.join(station_id)
    if constituent == 'Flow' or constituent == 'flow':
        #plt.title('Flow Duration Curve \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
        ax.set_title('Flow Duration Curve \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    else: 
        #plt.title(f'{constituent} Exceedance Plot \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
        ax.set_title(f'{constituent} Exceedance Plot \n Station(s): ' + stations, **CFONT, fontsize = 12, weight = 'bold')
    # Add x label
    ax.set_xlabel('Exceedance Percentage', **CFONT, fontsize = 10)
    # Add y label
    ax.set_ylabel(f'{constituent} ({units})', **CFONT, fontsize = 10) 
    # Add legend
    ax.legend(handles=[obs, sim], fontsize = 9)
    # Set legend font to Calibri
    L = ax.legend()
    plt.setp(L.texts, family = 'Calibri')
    # Set the font and font size for the tick labels
    plt.xticks(**CFONT, fontsize = 9)
    plt.yticks(**CFONT, fontsize = 9)
    # Add axis gridlines
    ax.grid(which='major', axis='both', linewidth=0.3)
    ax.grid(which='minor', axis='y', linewidth=0.3)
    # Convert y axis to log scale
    ax.set_yscale('log')
    # Set the x axis minimum to 0
    ax.set_xlim(left=0)
    
    # Save figure, if given a folder or file path in argument save_path
    if save_path is not None:
        save_path = Path(save_path)
        if save_path.is_dir():
            filepath = save_path.joinpath(constituent + '_FDCexceedPoints_' + stations + '.png')
            fig.savefig(filepath, bbox_inches='tight')
        else:
            filepath = save_path
            fig.savefig(filepath, bbox_inches='tight')

    return fig, ax


def _exceedence(df):
        
    df['simulated rank'] = df.loc[:,'simulated'].rank(method = 'average', ascending = False)
    df['simulated exceed %'] =  df.loc[:,'simulated rank'] / (len(df) + 1) * 100
    df['observed rank'] =  df.loc[:,'observed'].rank(method = 'average', ascending = False)
    df['observed exceed %'] =  df.loc[:,'observed rank'] / (len(df) + 1) * 100

    return df


# -*- coding: utf-8 -*-
"""
Created on Wed Jun 15 15:21:35 2022

@author: mfratki
"""
from mpcaHydro.data_manager import dataManager
from pyhspf.wdmReader import readWDM
from pyhspf.uci import UCI
from pyhcal.repository import Repository

import numpy as np
import pandas as pd
from pathlib import Path
import subprocess


def create_calibration_project(model_name,project_location,download_station_data = True,run_model = True,convert_wdms = True):
    project = Builder(model_name)
    project.copy(project_location,model_name)
    project.load_uci()
    project.format_uci()
    project.uci.build_targets().to_csv(project.project_path.joinpath('targets.csv'))
    if convert_wdms: project.convert_wdms()
    if download_station_data: project.download_station_data()
    if run_model: project.run_model()
   
    
    
    

class Builder():
    
    def __init__(self,model_name):
        self.repository = Repository(model_name)
        self.project_path = None
        self.project_name = None
        self.new_uci = None
        self.uci = None
        self.dm = None
    
    def valid_models():
        return Repository.valid_models()
    
    def set_project_path(self,project_path):
        self.project_path = Path(project_path)
        self.project_name = Path(project_path).name
        #self.new_uci = self.project_path.joinpath('model','_'.join([self.project_name,'0.uci']))
        #self.uci = UCI(self.project_path.joinpath('model','.'.join([self.project_name,'uci'])))
        
    def copy(self,project_location,project_name):
        self.project_path = Path(project_location).joinpath(project_name)
        self.project_name = project_name
        self.repository.copy(self.project_path)
        self.dm = dataManager(self.project_path.joinpath('data'))

    
    def load_uci(self):
        self.new_uci = self.project_path.joinpath('model','_'.join([self.project_name,'0.uci']))
        self.uci = UCI(self.project_path.joinpath('model','.'.join([self.project_name,'uci'])))

    def process(self):
        self.load_uci()
        self.format_uci()
        self.download_station_data()
        self.uci.build_targets().to_csv(self.project_path.joinpath('targets.csv'))

        # if not self.project_path.joinpath('model',self.project_name + '_0-0.hbn').exists():
        #     self.run_model()
        
    #TODO move to UCI class
    
    def format_uci(self,calibration_reaches = None):
        
        
        setup_files(self.uci,self.project_name,run = 0)
        setup_geninfo(self.uci)
        
        if calibration_reaches is None:
            calibration_reaches = self.repository.modl_db.loc[self.repository.modl_db['wplmn_flag'] == 1]['opnids'].str.split(',').to_list()
            calibration_reaches = [abs(int(j)) for i in calibration_reaches for j in i]        
        #calibration_reaches = self.uci.network.station_order(calibration_reaches)[-1][0]
        
        
        self.uci.initialize(name = self.project_name + '_0')
        # for reaches in self._MODL_DB['opnids'].str.split('+').to_list():
        #     [calibration_reaches.append(int(reach)) for reach in reaches if ~pd.isna(reach)]
            
        setup_binaryinfo(self.uci,reach_ids = calibration_reaches)
        setup_qualid(self.uci)
        self.uci.write(self.new_uci)
        
        # Download observation data
        # Sources/Databases WISKI and EQUIS (DELTA databases?)
    #TODO: use a single WISKI etl script for csg and wplmn data
    
    
    def download_station_data(self):#:,station_id,source_id):
        stations = self.repository.modl_db.drop_duplicates(subset = ['source','station_id'])
        if not stations.empty:
            for index, row in stations.iterrows():
                assert(row['source'].lower() in ['wiski','equis','swd'])
                station_id = row['station_id']
                self.dm.download_station_data(station_id,row['source'].lower())
            
    
    def convert_wdms(self):
        copy_path = Path(self.project_path.joinpath('model'))
        wdm_files = [copy_path.joinpath(wdm_file.name) for wdm_file in self.repository.wdm_files]
        for wdm_file in wdm_files:
            readWDM(wdm_file,
                    copy_path.joinpath(wdm_file.name.replace('.wdm','.hdf5').replace('.WDM','hdf5')))
        
    def run_model(self):
        # Run the uci file
        winHSPF = str(Path(__file__).resolve().parent.parent) + '\\bin\\WinHSPFLt\\WinHspfLt.exe'     
        subprocess.run([winHSPF,self.new_uci]) #, stdout=subprocess.PIPE, creationflags=0x08000000)    
        



### functions for setting up the UCI file properly 
def setup(uci,name,run = 0,reach_ids = None,n = 1,time_step = 3):
    uci = setup_files(uci,name,run,n)
    uci = setup_geninfo(uci)
    uci = setup_binaryinfo(uci,reach_ids = reach_ids)
    uci = setup_qualid(uci)


#def setup_files(uci,name,run = 0,n=5): # Add as method to a Files class?
#    # Update paths in Files block
    
def setup_files(uci,name,run,n = 5):
    table = uci.table('FILES',drop_comments = False)
    for index, row in table.iterrows():
        filename = Path(row['FILENAME'])
        if filename.suffix in ['.wdm','.ech','.out']:
            table.loc[index,'FILENAME'] = filename.name
        if filename.suffix in ['.hbn']:
            table.loc[index,'FILENAME'] = filename.name
        if filename.suffix in ['.plt']:
            table.drop(index,inplace = True)
            
    # Get new binary number and create new BINO rows
    bino_nums = []
    invalid = table['UNIT'].values
    for num in range(15,100):
        if num not in invalid:
            bino_nums.append(num)
        if len(bino_nums) == n:
            break
    binary_names = ['_'.join([name,str(run)])+ '-' + str(num) + '.hbn' for num in range(len( bino_nums))]
    rows = [['BINO',bino_num,binary_name,''] for bino_num,binary_name in zip(bino_nums,binary_names)]
    rows = pd.DataFrame(rows, columns = table.columns).astype({'FTYPE':'string','UNIT':'Int64','FILENAME':'string','comments':'string'} )
    # Drop old BINO rows and insert new BINO rows
    table = table.loc[table['FTYPE'] != 'BINO'].reset_index(drop=True)
    rows = pd.DataFrame(rows, columns = table.columns).astype(table.dtypes) #{'FTYPE':'string','UNIT':'Int64','FILENAME':'string','comments':'string'} )
    table = pd.concat([table,rows])
    table.reset_index(drop=True,inplace=True)
    
    # Update table in the uci
    uci.replace_table(table,'FILES')
    
    return uci

def setup_geninfo(uci):
    # Initialize Gen-Info
    bino_nums = uci.table('FILES').set_index('FTYPE').loc['BINO','UNIT'].tolist()
    if isinstance(bino_nums,int): #Pands is poorly designed. Why would tolist not return a goddamn list...?
        bino_nums = [bino_nums]
  
    #opnids = uci.table(operation,'GEN-INFO').index
    for operation in ['RCHRES','PERLND','IMPLND']:
        opnids = np.array_split(uci.table(operation,'GEN-INFO').index.to_list(),len(bino_nums))
        
        for opnid,bino_num in zip(opnids,bino_nums):
            if operation == 'RCHRES': #TODO convert BUNITE to BUNIT1 to get rid of this if statement
                uci.update_table(bino_num,'RCHRES','GEN-INFO',0,opnids = opnid,columns = 'BUNITE',operator = 'set')
            else:
                uci.update_table(bino_num,operation,'GEN-INFO',0,opnids = opnid,columns = 'BUNIT1',operator = 'set')
    return uci     

def setup_binaryinfo(uci,default_output = 4,reach_ids = None):
    # Initialize Binary-Info
    uci.update_table(default_output,'PERLND','BINARY-INFO',0,
                     columns = ['AIRTPR', 'SNOWPR', 'PWATPR', 'SEDPR', 'PSTPR', 'PWGPR', 'PQALPR','MSTLPR', 'PESTPR', 'NITRPR', 'PHOSPR', 'TRACPR'],
                     operator = 'set')
    uci.update_table(default_output,'IMPLND','BINARY-INFO',0,
                     columns = ['ATMPPR', 'SNOWPR', 'IWATPR', 'SLDPR', 'IWGPR', 'IQALPR'],
                     operator = 'set')
    uci.update_table(default_output,'RCHRES','BINARY-INFO',0, 
                     columns = ['HYDRPR', 'ADCAPR', 'CONSPR', 'HEATPR', 'SEDPR', 'GQLPR', 'OXRXPR', 'NUTRPR', 'PLNKPR', 'PHCBPR'],
                     operator = 'set')
        
    uci.update_table(default_output,'PERLND','BINARY-INFO',0,columns = ['SNOWPR','SEDPR','PWATPR','PQALPR'],operator = 'set')
    uci.update_table(default_output,'IMPLND','BINARY-INFO',0,columns = ['SNOWPR','IWATPR','SLDPR','IQALPR'],operator = 'set')
    uci.update_table(default_output,'RCHRES','BINARY-INFO',0,columns = ['HYDRPR','SEDPR','HEATPR','OXRXPR','NUTRPR','PLNKPR'],operator = 'set')
    if reach_ids is not None:
        uci.update_table(3,'RCHRES','BINARY-INFO',0,columns = ['SEDPR','OXRXPR','NUTRPR','PLNKPR'],opnids = reach_ids,operator = 'set')
        uci.update_table(2,'RCHRES','BINARY-INFO',0,columns = ['HEATPR','HYDRPR'],opnids = reach_ids,operator = 'set')
    return uci     


def setup_qualid(uci):
    #### Standardize QUAL-ID Names
    # Perlands
    uci.update_table('NH3+NH4','PERLND','QUAL-PROPS',0,columns = 'QUALID',operator = 'set')
    uci.update_table('NO3','PERLND','QUAL-PROPS',1,columns = 'QUALID',operator = 'set')
    uci.update_table('ORTHO P','PERLND','QUAL-PROPS',2,columns = 'QUALID',operator = 'set')
    uci.update_table('BOD','PERLND','QUAL-PROPS',3,columns = 'QUALID',operator = 'set')
    
    # Implands
    uci.update_table('NH3+NH4','IMPLND','QUAL-PROPS',0,columns = 'QUALID',operator = 'set')
    uci.update_table('NO3','IMPLND','QUAL-PROPS',1,columns = 'QUALID',operator = 'set')
    uci.update_table('ORTHO P','IMPLND','QUAL-PROPS',2,columns = 'QUALID',operator = 'set')
    uci.update_table('BOD','IMPLND','QUAL-PROPS',3,columns = 'QUALID',operator = 'set')
    return uci     



# def build_targets(uci,reach_ids,wplmn_map,wplmn_path,wplmn = None):
#     targets = uci.lc_info()
#     uci_names = targets['uci_name'].values
#     if wplmn is None:
#         wplmn = get_yields(reach_ids,uci,wplmn_path)
   
#     npsl_names = [wplmn_map[name] for name in targets['uci_name']]
#     targets['npsl_name'] = npsl_names #targets['npsl_name'] = npsl_names
    
    
    
#     npsl_rates = pd.concat([uci.npsl_targets.loc[name].mean() for name in npsl_names],axis = 1).transpose()
#     npsl_rates['uci_name'] = uci_names
#     npsl_rates = npsl_rates.melt(id_vars = 'uci_name',var_name = 'parameter', value_name = 'npsl_yield')
    
#     targets = npsl_rates.merge(targets,on = 'uci_name')
#     targets = targets.set_index(['uci_name','parameter'])
    
#     # normalize by dominant landcover of the model
#     dom_lc_name =  targets.index.get_level_values(0)[np.argmax(targets['area'])]
#     targets['npsl_yield_ratio'] = np.nan
#     for parameter in ['TSS','TKN','N','OP','BOD']:
#         values = targets.loc[:,parameter,:]['npsl_yield']/targets.loc[dom_lc_name,parameter]['npsl_yield']
#         targets.loc[values.index,'npsl_yield_ratio'] = values
    
#     # landcover area ratio compared to total area of watershed
#     targets['lc_area_ratio']= targets['area']/np.sum(targets.loc[:,'TSS',:]['area'])    
    
    
#     targets['wplmn_yield'] = np.nan
#     for parameter in ['TSS','TKN','N','OP','BOD']:
#         targets.loc[targets.loc[:,parameter,:].index,'wplmn_yield'] = wplmn[parameter]
      
#     # interative solution to determine target landcover loading rates
#     targets['lc_yield'] = np.nan
#     threshold = .05
#     for parameter in ['TSS','TKN','N','OP','BOD']:
#         print(parameter)
#         diff = 1
#         scaling_factor = 1
#         new_factor = 1
#         while diff > threshold:
#             scaling_factor = new_factor
#             lc_yield =  (targets['lc_area_ratio']*targets['npsl_yield_ratio']*scaling_factor).loc[:,parameter,:]
#             new_factor = wplmn[parameter]/np.sum(lc_yield)*scaling_factor
#             diff = np.abs((np.nansum(lc_yield)-wplmn[parameter])/wplmn[parameter])
#         print(scaling_factor)
#         values = targets.loc[targets.loc[:,parameter,:].index,'npsl_yield_ratio']*scaling_factor
#         targets.loc[values.index,'lc_yield'] = values.values
    

#     targets = uci.lc_info().set_index('uci_name').join(targets.unstack()['lc_yield'])
#     targets['dom_lc'] = np.nan
#     targets.loc[targets.index[np.argmax(targets['area'])],'dom_lc'] = 1
#     return targets

# def get_yields(reach_ids,uci,wplmn_path):
#     loadNetwork = dm.loadNetwork(wplmn_path,reach_ids)
    
#     schematic = uci.table('SCHEMATIC')
#     schematic = schematic.astype({'TVOLNO': int, "SVOLNO": int, 'AFACTR':float})
#     schematic = schematic[(schematic['SVOL'] == 'PERLND')| (schematic['SVOL'] == 'IMPLND')]
#     schematic = schematic[schematic['TVOL'] == 'RCHRES']
#     reaches = np.concatenate([nu.get_opnids(uci,reach_id)[0] for reach_id in reach_ids])
#     area = np.sum([schematic['AFACTR'][schematic['TVOLNO'] == reach].sum() for reach in reaches])
    
    
#     # wplmn2 = {'TSS': 18.01,
#     #           'TP': 0.23,
#     #           'OP':0.082,
#     #           'TKN':1.921,
#     #           'N':2.304}
#     # wplmn2['BOD'] = 50*(wplmn['TP'] - wplmn['OP'])
#     # wplmn2['TKN'] = wplmn['TKN'] - wplmn['BOD']*.144

#     wplmn = {'TSS': 0,
#               'TP': 0,
#               'OP':0,
#               'TKN':0,
#               'N':0}
    
#     for key in wplmn.keys():
#         if key == 'TSS':
#             wplmn[key] = np.mean(loadNetwork.get_timeseries(key,'Load').groupby('year').sum()['Ts Value']/area*2000)
#         else:
#             wplmn[key] = np.mean(loadNetwork.get_timeseries(key,'Load').groupby('year').sum()['Ts Value']/area)

        
#     wplmn['BOD'] = 50*(wplmn['TP'] - wplmn['OP'])
#     wplmn['TKN'] = wplmn['TKN'] - wplmn['BOD']*.144
#     return wplmn

# def lc_info(uci):
#     uci.get_metzones2()
#     geninfo = uci.table('PERLND','GEN-INFO')  
#     targets = uci.opnid_dict['PERLND'].loc[:,['LSID','landcover']] #.drop_duplicates(subset = 'landcover').loc[:,['LSID','landcover']].reset_index(drop = True)
#     targets.columns = ['LSID','lc_number']
#     schematic = uci.table('SCHEMATIC')
#     schematic = schematic.astype({'TVOLNO': int, "SVOLNO": int, 'AFACTR':float})
#     schematic = schematic[(schematic['SVOL'] == 'PERLND')]
#     schematic = schematic[(schematic['TVOL'] == 'PERLND') | (schematic['TVOL'] == 'IMPLND') | (schematic['TVOL'] == 'RCHRES')]
#     areas = []
#     for lc_number in targets['lc_number'].unique():
#         areas.append(np.sum([schematic['AFACTR'][schematic['SVOLNO'] == perland].sum() for perland in targets.index[targets['lc_number'] == lc_number]]))
#     areas = np.array(areas)
    
    
#     lc_number = targets['lc_number'].drop_duplicates()
#     uci_names = geninfo.loc[targets['lc_number'].drop_duplicates().index]['LSID']
#     targets = pd.DataFrame([uci_names.values,lc_number.values,areas]).transpose()
#     targets.columns = ['uci_name','lc_number','area']
#     targets['npsl_name'] = ''
    
#     targets[['TSS','N','TKN','OP','BOD']] = ''
    
#     targets['dom_lc'] = ''
#     targets.loc[targets['area'].astype('float').argmax(),'dom_lc'] = 1
#     return targets





#WPLMN = pd.read_csv()
# HUC_DIRECTORY = pd.read_csv('C:/Users/mfratki/Documents/Github/hspf_tools/calibrator/HUC_Names.csv',dtype = {'USGS HUC-8':'string',
#                                                                                                              'USGS HUC-6':'string',
#                                                                                                              'USGS HUC-4':'string',
#                                                                                                              'USGS HUC-2':'string'})
#MODL_DB = pd.read_csv('C:/Users/mfratki/Documents/Github/hspf_tools/calibrator/WPLMN.csv',dtype = {'stn_HUC12':'string'})


# REPO = Path('X:\Databases2\Water_Quality\Watershed_Modeling\MPCA_HSPF_Model_Repository') #could point to the github website if it becomes public
# huc8_id = '10170204'
# huc6_name = HUC_DIRECTORY.loc[HUC_DIRECTORY['USGS HUC-6'] == huc8_id[:6],'Repository_HUC6 Name'].iloc[0]
# model_name = HUC_DIRECTORY.loc[HUC_DIRECTORY['USGS HUC-8'] == huc8_id[:8],'Repository_HUC8 Name'].iloc[0]
# repo_folder = [item for item in REPO.joinpath('_'.join([huc6_name,huc8_id[:6]])).iterdir() if item.name.startswith(model_name)]
# repo_folder = repo_folder[0]    

# stations = gpd.read_file('C:/Users/mfratki/Documents/Projects/MODL Database/stations_WISKI2.gpkg')
# stations.dropna(subset='stn_HUC12',inplace=True)
# stations['opnids'] = stations['opnids'].str.replace('-','+')
# HUC_DIRECTORY = pd.read_csv('C:/Users/mfratki/Documents/Github/hspf_tools/calibrator/HUC_Names.csv',dtype = {'USGS HUC-8':'string',
#                                                                                                               'USGS HUC-6':'string',
#                                                                                                               'USGS HUC-4':'string',
#                                                                                                               'USGS HUC-2':'string'})
# test = pd.merge(wiski, HUC_DIRECTORY, left_on=  'HUC8',
#                     right_on= 'USGS HUC-8', 
#                     how = 'left')
# class Builder():
#     # HUC_DIRECTORY = pd.read_csv('C:/Users/mfratki/Documents/Github/hspf_tools/calibrator/HUC_Names.csv',dtype = {'USGS HUC-8':'string',
#     #                                                                                                               'USGS HUC-6':'string',
#     #                                                                                                               'USGS HUC-4':'string',
#     #                                                                                                               'USGS HUC-2':'string'})
#     # MODL_DB = pd.read_csv('C:/Users/mfratki/Documents/Github/hspf_tools/calibrator/WPLMN.csv',dtype = {'stn_HUC12':'string',
#     #                                                                                                    'USGS HUC-8':'string',
#     #                                                                                                    'USGS HUC-6':'string',
#     #                                                                                                    'USGS HUC-4':'string',
#     #                                                                                                    'USGS HUC-2':'string'}

#     # MODL_DB = pd.read_csv(Path(__file__).parent/'stations_WISKI.csv',dtype = {'stn_HUC12':'string',              
#     #                                                                           'USGS HUC-8':'string',
#     #                                                                           'USGS HUC-6':'string',
#     #                                                                           'USGS HUC-2':'string',
#     #                                                                           'USGS HUC-4':'string'})
#     # MODL_DB = pd.read_csv('C:/Users/mfratki/Documents/Projects/MODL Database/stations_WISKI.csv', dtype = {'stn_HUC12':'string',              
#     #                                                                                                        'USGS HUC-8':'string',
#     #                                                                                                        'USGS HUC-6':'string',
#     #                                                                                                        'USGS HUC-4':'string',
#     #                                                                                                        'USGS HUC-2':'string'})                                                                                                       

#     MODL_DB = gpd.read_file('C:/Users/mfratki/Documents/Projects/MODL Database/gis/stations_csg.gpkg').dropna(subset='opnids')
#     #MODL_DB = gpd.read_file('C:/Users/mfratki/Documents/Projects/MODL Database/stations_WISKI2.gpkg')  #.dropna(subset='opnids')
#     #MODL_DB['opnids'] = MODL_DB['opnids'].str.replace('-','+')    
#     MODL_DB = MODL_DB.dropna(subset='opnids')
#     MODL_DB = MODL_DB.loc[~(MODL_DB['opnids'] == -1)]
#     REPO = Path('X:\Databases2\Water_Quality\Watershed_Modeling\MPCA_HSPF_Model_Repository') #could point to the github website if it becomes public
    
#     @classmethod
#     def valid_models(self):
#         return list(self.MODL_DB['Repository_HUC8 Name'].dropna()) #replace('NO MODEL',pd.NA).dropna())

                
#     def __init__(self,model_name):# = None, huc8_id = None)
#         if model_name not in self.valid_models():
#             print('Please provide a valid model name (see .valid_models)')
#             return
            
#         self.model_name = model_name
#         #self._HUC_DIRECTORY = self.HUC_DIRECTORY.loc[self.HUC_DIRECTORY['Repository_HUC8 Name'] == self.model_name]
#         #self.huc6_name = self._HUC_DIRECTORY['Repository_HUC6 Name'].iloc[0]
#         #self.huc8_id = self._HUC_DIRECTORY['USGS HUC-8'].iloc[0] #HUC_DIRECTORY.loc[HUC_DIRECTORY['USGS HUC-6'] == huc8_id[:6],'Repository_HUC6 Name'].iloc[0]
        
#         self._MODL_DB = self.MODL_DB.loc[self.MODL_DB['Repository_HUC8 Name'] == self.model_name]
#         self.huc6_name = self._MODL_DB['Repository_HUC6 Name'].iloc[0]
#         self.huc6_id = self._MODL_DB['USGS HUC-6'].iloc[0]
#         self.huc8_ids = self._MODL_DB['USGS HUC-8'].unique()
#         self.repo_folder = [item for item in self.REPO.joinpath('_'.join([self.huc6_name,self.huc6_id])).iterdir() if item.name.startswith(self.model_name)][0]        
#         self.uci_file = self.repo_folder.joinpath('HSPF','.'.join([self.model_name,'uci']))
#         self.wdm_files = [item for item in self.repo_folder.joinpath('HSPF').iterdir() if item.name.endswith('.wdm')]
    
#     def copy(self,project_location):


#         build_folders(self.model_name,project_location)
#         self.project_path = Path(project_location).joinpath(self.model_name)
#         self.new_uci = self.project_path.joinpath('model','_'.join([self.model_name,'0.uci']))
#         shutil.copyfile(self.uci_file, self.project_path.joinpath('model','.'.join([self.model_name,'uci'])))
#         self.uci = UCI(self.project_path.joinpath('model','.'.join([self.model_name,'uci'])))
#         for wdm_file in self.wdm_files:
#             shutil.copyfile(wdm_file,self.project_path.joinpath('model',Path(wdm_file).name))
    
        
#         self.format_uci()
#         self.download_wplmn()
#         self.download_csg()
#         self.build_lc_targets()
#         if not self.project_path.joinpath('model',self.model_name + '_0-0.hbn').exists():
#             self.run_model()
#         self._MODL_DB.to_csv(self.project_path.joinpath('_'.join([self.model_name,'MODL_DB.csv'])))
        
#     def format_uci(self):
        
#         setup_files(self.uci,self.model_name,run = 0)
#         setup_geninfo(self.uci)
        
#         calibration_reaches = self._MODL_DB['opnids'].astype('int').to_list()
#         # for reaches in self._MODL_DB['opnids'].str.split('+').to_list():
#         #     [calibration_reaches.append(int(reach)) for reach in reaches if ~pd.isna(reach)]
            
#         setup_binaryinfo(self.uci,reach_ids = calibration_reaches)
#         setup_qualid(self.uci)
#         self.uci.write(self.new_uci)
        
#         # Download observation data
#         # Sources/Databases WISKI and EQUIS (DELTA databases?)
#     #TODO: use a single WISKI etl script for csg and wplmn data
#     def download_csg(self):
#         if self.project_path.joinpath('csg.csv').exists():
#             print('CSG data already downloaded')
#         station_nos = (self._MODL_DB.loc[self._MODL_DB['WPLMN'] == 0,'station_no'].unique())
#         if len(station_nos) == 0:
#             print('No CSG stations linked to Watershed yet')
#             return
#         data = pd.concat([etlCSG.download(station_no).pipe(etlCSG.transform) for station_no in station_nos])
#         #reach_map = dict(self._MODL_DB.loc[self._MODL_DB['station_no'].isin(station_nos),['station_no','opnids']].values)
#         #data['reach_id'] = data['station_no'].map(reach_map)
#         etlCSG.load(data,self.project_path.joinpath('csg.csv'))
             
#     def download_wplmn(self):
#         if self.project_path.joinpath('wplmn.csv').exists():
#             print('WPLMN data already downloaded')
#         station_nos = list(self._MODL_DB.loc[self._MODL_DB['WPLMN'] == 1,'station_no'].unique())
#         data = pd.concat([etlWPLMN.download(station_no).pipe(etlWPLMN.transform) for station_no in station_nos])
#         #reach_map = dict(self._MODL_DB.loc[self._MODL_DB['station_no'].isin(station_nos),['station_no','opnids']].values)
#         #data['reach_id'] = data['Station number'].map(reach_map)
#         etlCSG.load(data,self.project_path.joinpath('wplmn.csv'))
    
#     def download_swd(self):
#         if self.project_path.joinpath('swd.csv').exists():
#             print('SWD data already downloaded')
#         station_nos = list(self._MODL_DB.loc[self._MODL_DB['source'] == 'swd','station_no'].unique())
#         data = pd.concat([etlSWD.download(station_no).pipe(etlSWD.transform) for station_no in station_nos])
#         etlSWD.load(data,self.project_path.joinpath('swd.csv'))

#         # Create landcover targets spreadsheet
#     def build_lc_targets(self):
#         lc_info(self.uci).to_csv(self.project_path.joinpath('targets.csv'))
        
#     def run_model(self):
#         # Run the uci file
#         winHSPF = str(Path(__file__).resolve().parent.parent) + '\\bin\\WinHSPFLt\\WinHspfLt.exe'     
#         subprocess.run([winHSPF,self.new_uci]) #, stdout=subprocess.PIPE, creationflags=0x08000000)    
        

#         # if (model_name is None) & (huc8_id is None):
#         #     print('Please provide Either a HUC8 ID or a valid model name. To see valid model names see .valid_models')
#         #     return
#         #if model_name is None:
#         # else:
#         #     self._HUC_DIRECTORY = self.HUC_DIRECTORY.loc[self.HUC_DIRECTORY['Repository_HUC8 Name',:]
#         #     self.huc6_name = self._HUC_DIRECTORY.loc[0,'Repository_HUC6 Name']
#         #     self.model_name = self._HUC_DIRECTORY.loc[0,'Repository_HUC8 Name']
#         #     self.repo_folder = 





# load original uci

# def setup_files(uci,project_name): # Add as method to a Files class?
#     # Update paths in Files block
#     table = uci.table('FILES',drop_comments = False)
#     for index, row in table.iterrows():
#         name = Path(row['FILENAME'])
#         if name.suffix in ['ech','out','wdm']:
#             table.loc[index,'FILENAME'] = name
#         if name.suffix in ['hbn']:
#             table.loc[index,'FILENAME'] = name
#         if name.suffix in ['plt']:
#             table.drop(index)
    
#     table = table.set_index(table.columns[0])


#     # Rename hbn in Files block
#     #TODO: dynamic size management of hbn files
#     new_bino = table.loc['BINO'][~table.loc['BINO'].index.duplicated()]
#     table = table.drop('BINO',axis=0)
#     table = table.append(new_bino)
#     table.loc['BINO','FILENAME'] = project_name + '_0.hbn'
#     table = table.reset_index()
#     uci.replace_table(table,'FILES')        
#     return uci     


# def setup_geninfo(uci):
#     # Initialize Gen-Info
#     bino_num = uci.table('FILES').set_index('FTYPE').loc['BINO','UNIT']
#     uci.update_table(bino_num,'PERLND','GEN-INFO',0,columns = 'BUNIT1',operator = 'set')
#     uci.update_table(bino_num,'RCHRES','GEN-INFO',0,columns = 'BUNITE',operator = 'set')
#     uci.update_table(bino_num,'IMPLND','GEN-INFO',0,columns = 'BUNIT1',operator = 'set')
#     return uci     

# def setup_binaryinfo(uci,reach_ids = None):
#     # Initialize Binary-Info
#     uci.update_table(4,'PERLND','BINARY-INFO',0,columns = ['SNOWPR','SEDPR','PWATPR','PQALPR'],operator = 'set')
#     uci.update_table(4,'IMPLND','BINARY-INFO',0,columns = ['SNOWPR','IWATPR','SLDPR','IQALPR'],operator = 'set')
#     uci.update_table(4,'RCHRES','BINARY-INFO',0,columns = ['HYDRPR','SEDPR','HEATPR','OXRXPR','NUTRPR','PLNKPR'],operator = 'set')
#     if reach_ids is not None:
#         uci.update_table(3,'RCHRES','BINARY-INFO',0,columns = ['HYDRPR','SEDPR','HEATPR','OXRXPR','NUTRPR','PLNKPR'],opnids = reach_ids,operator = 'set')
#     return uci     

# def setup_qualid(uci):
#     #### Standardize QUAL-ID Names
#     # Perlands
#     uci.update_table('NH3+NH4','PERLND','QUAL-PROPS',0,columns = 'QUALID',operator = 'set')
#     uci.update_table('NO3','PERLND','QUAL-PROPS',1,columns = 'QUALID',operator = 'set')
#     uci.update_table('ORTHO P','PERLND','QUAL-PROPS',2,columns = 'QUALID',operator = 'set')
#     uci.update_table('BOD','PERLND','QUAL-PROPS',3,columns = 'QUALID',operator = 'set')
    
#     # Implands
#     uci.update_table('NH3+NH4','IMPLND','QUAL-PROPS',0,columns = 'QUALID',operator = 'set')
#     uci.update_table('NO3','IMPLND','QUAL-PROPS',1,columns = 'QUALID',operator = 'set')
#     uci.update_table('ORTHO P','IMPLND','QUAL-PROPS',2,columns = 'QUALID',operator = 'set')
#     uci.update_table('BOD','IMPLND','QUAL-PROPS',3,columns = 'QUALID',operator = 'set')
#     return uci     









# -*- coding: utf-8 -*-
"""
Created on Thu May  1 09:51:51 2025

@author: mfratki
"""
#import sqlite3
from pathlib import Path
import geopandas as gpd
import pandas as pd
#from hspf_tools.calibrator import etlWISKI, etlSWD


#stations_wiski = gpd.read_file('C:/Users/mfratki/Documents/GitHub/pyhcal/src/pyhcal/data/stations_wiski.gpkg')


stations_wiski = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_wiski.gpkg')).dropna(subset='opnids')[['station_id','true_opnid','opnids','comments','modeled','repository_name','wplmn_flag']]
stations_wiski['source'] = 'wiski'
stations_equis = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_EQUIS.gpkg')).dropna(subset='opnids')[['id_code','true_opnid','opnids','comments','modeled','repository_name']]
stations_equis['source'] = 'equis'
stations_equis['wplmn_flag'] = 0
stations_equis = stations_equis.rename(columns = {'id_code':'station_id'})


MODL_DB = pd.concat([stations_wiski,stations_equis])

database  = """
    -- Stations/Locations table
    CREATE TABLE IF NOT EXISTS Station (
        stationPK INTEGER PRIMARY KEY AUTOINCREMENT,
        reachPK INTEGER REFERENCES Reach(reachPK),
        stationID TEXT NOT NULL,
        stationName TEXT,
        stationOrigin TEXT NOT NULL,
        latitude REAL,
        longitude REAL,
        stationType TEXT,
        UNIQUE(stationID, stationOrigin)
    );
    
    -- Station Associations table
    CREATE TABLE IF NOT EXISTS StationAssociations (
        stationPK INTEGER REFERENCES Station(stationPK),
        associationPK INTEGER REFERENCES Station(stationPK)
    );

    -- Station Aliases table
    CREATE TABLE IF NOT EXISTS StationAliases (
        stationPK INTEGER NOT NULL,
        aliasPK INTEGER NOT NULL,
        FOREIGN KEY (stationPK) REFERENCES Station(stationPK),
        FOREIGN KEY (aliasPK) REFERENCES Station(stationPK)
    );
    
    CREATE TABLE Reach (
        reachPK INTEGER PRIMARY KEY,
        modelName TEXT NOT NULL,
        reachID INTEGER NOT NULL,
        drainageArea FLOAT 
    );
    
    CREATE TABLE Outlet (
        outletPK INTEGER PRIMARY KEY,
        outletName TEXT
    );
    
    -- Outlet-Station Associations table
    CREATE TABLE IF NOT EXISTS StationAssociations (
        outletPK INTEGER NOT NULL REFERENCES Outlet(outletPK),
        stationPK  INTEGER NOT NULL REFERENCES Station(reachPK)
    );
    
    -- Outlet-Reach Associations table
    CREATE TABLE IF NOT EXISTS StationAssociations (
        outletPK INTEGER NOT NULL REFERENCES Outlet(outletPK),
        reachPK  INTEGER NOT NULL REFERENCES Station(reachPK)
        exclude INTEGER NOT NULL
    );"""
    
    
#row = modl_db.MODL_DB.iloc[0]

#info = etlWISKI.info(row['station_id'])

#modl_db.MODL_DB.query('source == "equis"')

# outlet_dict = {'stations': {'wiski': ['E66050001'],
#                'equis': ['S002-118']},
#                'reaches': {'Clearwater': [650]}
                      



# station_ids = ['S002-118']
# #station_ids = ['E66050001']
# reach_ids = [650]
# flow_station_ids =  ['E66050001']



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 27 09:16:30 2024

@author: mfratki
"""

import pandas as pd
from pyhcal.modl_db import MODL_DB
from pathlib import Path
import shutil


DEFUALT_REPOSITORY_PATH = Path('X:\Databases2\Water_Quality\Watershed_Modeling\MPCA_HSPF_Model_Repository') #could point to the github website if it becomes public?

class Repository():
    
    HUC_DIRECTORY = pd.read_csv(str(Path(__file__).resolve().parent/'data\\HUC_Names.csv'),dtype = {'USGS HUC-8':'string',
                                                                                                                  'USGS HUC-6':'string',
                                                                                                                  'USGS HUC-4':'string',
                                                                                                                  'USGS HUC-2':'string'})
    
    MODL_DB = MODL_DB

    
    

    @classmethod
    def valid_models(self):
        return sorted(list(set(self.HUC_DIRECTORY['Repository_HUC8 Name'].dropna().replace('NO MODEL',pd.NA).dropna().to_list())))
    
    
    def __init__(self,model_name,repository_path = DEFUALT_REPOSITORY_PATH):# = None, huc8_id = None)
        if model_name not in self.valid_models():
            print('Please provide a valid model name (see .valid_models)')
            return
        
        
        self.REPOSITORY_PATH = repository_path
        huc_directory = self.HUC_DIRECTORY.loc[self.HUC_DIRECTORY['Repository_HUC8 Name'] == model_name]
        self.modl_db = self.MODL_DB.loc[self.MODL_DB['repository_name'] == model_name]
        #self.modl_db  = pd.concat([self.MODL_DB.loc[self.MODL_DB['repository_name'].str.startswith(huc8_id,na=False)] for huc8_id in huc8_ids])        
        self.model_name = model_name
        self.huc8_ids = list(huc_directory['USGS HUC-8'])
        self.huc6_name = huc_directory['Repository_HUC6 Name'].iloc[0]
        self.huc6_id =  huc_directory['USGS HUC-6'].iloc[0]
        self.repo_folder = [item for item in self.REPOSITORY_PATH.joinpath('_'.join([self.huc6_name,self.huc6_id])).iterdir() if item.name.split('_')[0] == self.model_name][0]        
        self.uci_file = self.repo_folder.joinpath('HSPF','.'.join([self.model_name,'uci']))
        self.wdm_files = [item for item in self.repo_folder.joinpath('HSPF').iterdir() if (item.name.endswith('.wdm')) | (item.name.endswith('.WDM'))]
        self.shapefiles = {item.name.split('.')[0].split('_')[-1]:item for item in self.repo_folder.joinpath('GIS').iterdir() if (item.name.endswith('.shp')) | (item.name.endswith('.SHP'))}

    def copy(self,copy_path):
        copy_path = Path(copy_path)
        build_folders(copy_path)
        shutil.copyfile(self.uci_file, copy_path.joinpath('model','.'.join([self.model_name,'uci'])))
        for wdm_file in self.wdm_files:
            shutil.copyfile(wdm_file,copy_path.joinpath('model',Path(wdm_file).name))
        self.modl_db.to_csv(copy_path.joinpath('_'.join([self.model_name,'MODL_DB.csv'])))
        self.copy_shapefiles(copy_path.joinpath('gis'))

    def copy_wdm(self,copy_path):
        for wdm_file in self.wdm_files:
            shutil.copyfile(wdm_file,Path(copy_path).joinpath(Path(wdm_file).name))

    def copy_uci(self,copy_path):
        shutil.copyfile(self.uci_file, Path(copy_path).joinpath('.'.join([self.model_name,'uci'])))
 
    def copy_modl_db(self,copy_path):
        self.modl_db.to_csv(Path(copy_path).joinpath('_'.join([self.model_name,'MODL_DB.csv'])))

    def copy_shapefiles(self,copy_path):
        for k,shapefile in self.shapefiles.items():
            files = [file for file in shapefile.parent.iterdir() if file.stem == shapefile.stem]
            [shutil.copyfile(file,Path(copy_path).joinpath(Path(file).name)) for file in files]
            

    
def build_folders(trg_path):
    
    sub_folders = ['model',
                   'output',
                   'inputs',
                   'figures',
                   'gis',
                   'data']
    
    
    trg_path = Path(trg_path)
    #p = Path(build_path) 
    #trg_path = p.joinpath(project_name)
    if not trg_path.is_dir():
        trg_path.mkdir(parents=True)
    
    for path in sub_folders:
        if not trg_path.joinpath(path).is_dir():
            trg_path.joinpath(path).mkdir()

      
